<?php
// Encoding: UTF-8 (without BOM) (auto-detect: °°°°°)
require('_head.inc.php');

// ----- Handle form requests and status messages -----
if (UnbUi::IsFormSuccess())
{
	$newUser = UnbUi::GetFormHandler()->GetNewUser();
	$editedUser = UnbUi::GetFormHandler()->GetEditedUser();
	$anyUser = isset($newUser) ? $newUser : $editedUser;
	if ($newUser)
		UnbSession::SetVar('SuccessMessage', 'User “' . $newUser->GetDisplayNameOrId() . '” created');
	else if ($editedUser)
		UnbSession::SetVar('SuccessMessage', 'User “' . $editedUser->GetDisplayNameOrId() . '” saved');
	if ($_POST['continue_alter_user'])
		UnbUi::RedirectAndExit('create_user.php?userid=' . $anyUser->GetUserId());
	else
		UnbUi::RedirectAndExit('list_users.php');
}

UnbUi::BeginReport();
UnbUi::ReportSuccess();
UnbUi::ReportError();

// ----- Load the user -----
$alterUserId = isset($_REQUEST['userid']) ? intval($_REQUEST['userid']) : null;

if (isset($alterUserId))
{
	$alterUser = UnbUser::FindByUserId($alterUserId);
}

$error = false;
if (!isset($alterUser) && !UnbSession::IsAdministrator() ||
	isset($alterUser) && $alterUser->GetUserId() != UnbSession::GetSessionUserId() && !UnbSession::IsAdministrator())
{
	UnbUi::ReportError('Access denied');
	$error = true;
}
else if (isset($alterUserId) && !isset($alterUser))
{
	UnbUi::ReportError('Invalid user ID');
	$error = true;
}
UnbUi::EndReport();

if (!$error)
{
	if (isset($alterUser))
		echo '<h1>Alter user ' . unbqt(htmlspecialchars($alterUser->GetDisplayNameOrId())) . '</h1>';
	else
		echo '<h1>Create a new user</h1>';
	?>

	<?php UnbUi::BeginForm() ?>
	<?php
	if (isset($alterUser))
		UnbUi::FormUser($alterUser->GetUserId());
	else
		echo '<input type="hidden" name="action[]" value="create_user" />';
	?>

	<h2>General data</h2>

	<table cellspacing="0" class="grid first-10em">
	<tr>
		<td class="vspace">Display name:</td>
		<td><?php UnbUi::UserDisplayName($alterUser, null, '', !isset($alterUser)) ?></td>
		<td class="form-description">This name will be displayed above your messages.</td>
	</tr>

	<tr>
		<td colspan="3">
			<h2>Login information</h2>
		</td>
	</tr>
	<?php
	if (isset($alterUser) && $alterUser->GetLoginPassword() && !UnbSession::IsAdministrator())
	{
		?>
		<tr>
			<td class="vspace">Current password:</td>
			<td><?php UnbUi::UserCurrentPassword() ?></td>
			<td class="form-description">You need to enter your current password to change your login name or password.</td>
		</tr>
		<?php
	}
	if (isset($alterUser) && $alterUser->GetUserId() == UnbSession::GetSessionUserId() && !$alterUser->GetLoginPassword())
	{
		?>
		<tr>
			<td colspan="3">
				<p class="warningbox">
				<strong>Attention:</strong> You cannot log into this account unless you set a login name and password below.
				</p>
			</td>
		</tr>
		<?php
	}
	?>
	<tr>
		<td class="vspace">Login name:</td>
		<td><?php UnbUi::UserLoginName($alterUser, null, '', !isset($alterUser)) ?></td>
		<td class="form-description">The login name is used for logging into the system and is independent from the display name.<br />
			It should contain letters and numbers only, no special or local characters.</td>
	</tr>
	<tr>
		<td class="vspace"><?php echo isset($alterUser) && $alterUser->GetLoginPassword() ? 'Change' : 'Set' ?> password:</td>
		<td><?php UnbUi::UserLoginPassword($alterUser, null, '', !isset($alterUser)) ?></td>
		<td class="form-description">Keep this password in confidence and do not give it away to anybody.</td>
	</tr>
	<tr>
		<td class="vspace">Repeat password:</td>
		<td><?php UnbUi::UserRepeatPassword() ?></td>
		<td class="form-description">To avoid typing errors, enter the password for a second time.</td>
	</tr>

	<tr>
		<td colspan="3">
			<h2>Access</h2>
		</td>
	</tr>
	<tr>
		<td>Special privileges:</td>
		<?php
		if (UnbSession::IsAdministrator())
		{
			?>
			<td colspan="2">
				<?php UnbUi::UserAdministrator($alterUser) ?><br />
				<?php UnbUi::UserModerator($alterUser) ?><br />
				<?php UnbUi::UserTrusted($alterUser) ?><br />
			</td>
			<!-- td class="form-description">Administrator/moderator/trusted user status.</td -->
			<?php
		}
		else
		{
			?>
			<td colspan="2">
				<?php
				$statusStr = '';
				if ($alterUser->IsAdministrator())
					$statusStr .= ($statusStr ? ', ' : '') . 'Administrator';
				if ($alterUser->IsModerator())
					$statusStr .= ($statusStr ? ', ' : '') . 'Moderator';
				if ($alterUser->IsTrusted())
					$statusStr .= ($statusStr ? ', ' : '') . 'Trusted user';
				if ($statusStr)
					echo $statusStr;
				else
					echo '<em>(None)</em>';
				?>
			</td>
			<?php
		}
		?>
	</tr>
	<tr>
		<td>Additional keys:</td>
		<?php
		if (UnbSession::IsAdministrator())
		{
			?>
			<td><?php UnbUi::UserAdditionalKeysList($alterUser) ?></td>
			<td class="form-description">List of access keys that the user posesses, one per line by their display name.</td>
			<?php
		}
		else
		{
			?>
			<td colspan="2">
				<?php
				$additionalKeysList = UnbUi::GetUserAdditionalKeys($alterUser);
				if (strlen($additionalKeysList))
					echo nl2br(htmlspecialchars($additionalKeysList));
				else
					echo '<em>(None)</em>';
				?>
			</td>
			<?php
		}
		?>
	</tr>

	<?php
	if (isset($alterUser))
	{
		?>
		<tr>
			<td colspan="3">
				<h2>vCard entries</h2>
			</td>
		</tr>
		<?php
		foreach ($alterUser->GetVcardEntries() as $type => $entry)
		{
			?>
			<tr>
				<td><?php echo htmlspecialchars(unbtr('vcard:' . mb_strtolower($type), true)) ?>:</td>
				<td colspan="2"><?php UnbUi::UserAlterVcardEntry($type, $entry) ?></td>
			</tr>
			<?php
		}
		?>
		<tr>
			<td colspan="3">
				Add a new entry:
			</td>
		</tr>
		<tr>
			<td><?php UnbUi::UserVcardTypes(array(
				'Birthday',
				'EMail',
				'ICQ',
				'Jabber',
				'Location',
				'MSN',
				'RealName',
				'Slogan',
				'Website',
			)) ?></td>
			<td colspan="2"><?php UnbUi::UserAddVcardEntry() ?></td>
		</tr>
		<tr>
			<td colspan="2">
			</td>
			<td class="form-description">
				To add a new entry, select the entry type and enter its data. Save changes once for another entry.<br />
				<em>Public</em> is visible for everybody, <em>Internal</em> for all logged in users, <em>Confidential</em> only for trusted users and <em>Private</em> only for you.
			</td>
		</tr>
		<?php
	}
	?>

	<tr>
		<td></td>
		<td colspan="2">
			<br />
			<?php
			if (isset($alterUser))
			{
				echo '<input type="submit" name="continue_alter_user" class="icon13 ok" value="Save and continue" /> &nbsp; ';
				echo '<input type="submit" class="default icon13 ok" value="Save changes" /> &nbsp; ';
			}
			else
			{
				echo '<input type="submit" name="continue_alter_user" class="icon13 ok" value="Create and edit" /> &nbsp; ';
				echo '<input type="submit" class="default icon13 ok" value="Create user" /> &nbsp; ';
			}
			?>
			<a href="list_users.php" class="buttonlike icon13 back">Back to users list</a>
			<?php
			if (isset($alterUser) && UnbSession::IsAdministrator())
			{
				?>
				&nbsp; <a href="delete_user.php?userid=<?php echo $alterUser->GetUserId() ?>" class="buttonlike icon13 delete">Delete this user…</a>
				<?php
			}
			?>
		</td>
	</tr>
	</table>
	<?php UnbUi::EndForm() ?>

	<script type="text/javascript">//<![CDATA[
	$("#userDisplayNameTb").focus();
	//]]></script>

	<?php
}
require('_foot.inc.php');
?>