<?php
// Encoding: UTF-8 (without BOM) (auto-detect: °°°°°)
require('_head.inc.php');

$messageId = isset($_REQUEST['messageid']) ? intval($_REQUEST['messageid']) : null;
$messageRevisionId = isset($_REQUEST['messagerevisionid']) ? intval($_REQUEST['messagerevisionid']) : null;
$rootMessageId = isset($_REQUEST['root']) ? intval($_REQUEST['root']) : null;
$returnToMessageId = isset($_REQUEST['retmid']) ? intval($_REQUEST['retmid']) : null;

// ----- Handle form requests and status messages -----
if (UnbUi::IsFormSuccess())
{
	$reqMessageId = null;
	if (UnbUi::GetFormHandlerFields('moderate_messagerevision', 'success') !== null)
	{
		if ($_REQUEST['message_modstate'] == UnbMessageRevisionModerationState::Approved)
			UnbSession::SetVar('SuccessMessage', unbtrc('msg:message revision approved/unlocked'));
		if ($_REQUEST['message_modstate'] == UnbMessageRevisionModerationState::Locked)
			UnbSession::SetVar('SuccessMessage', unbtrc('msg:message revision locked'));
		$reqMessageId = UnbUi::GetFormHandler()->GetRequestedMessageRevision()->GetMessageId();
	}

	if (isset($returnToMessageId))
	{
		$messageId = $returnToMessageId;
		$messageRevisionId = null;
	}

	$params = array();
	$hash = null;
	if (isset($messageId))
	{
		$params['messageid'] = $messageId;
		$hash = 'message-' . $messageId;
	}
	if (isset($messageRevisionId))
	{
		$params['messagerevisionid'] = $messageRevisionId;
		if ($reqMessageId)
		{
			$hash = 'message-' . $reqMessageId;
		}
	}
	UnbUi::RedirectAndExit('view_message.php', $params, $hash);
}

UnbUi::BeginReport();
UnbUi::ReportSuccess();
UnbUi::ReportError();

// ----- Load the message revision -----
$error = false;
if (!$messageId && !$messageRevisionId)
{
	UnbUi::ReportError('No message selected.');
	$error = true;
}
else if (!UnbUi::GetMsgRev($messageId, $messageRevisionId, $msg, $rev, $err, true))
{
	UnbUi::ReportError($err);
	$error = true;
}
UnbUi::EndReport();

if (!$error)
{
	$markup = new UnbMarkup();
	$showAvatars = false;

	if ($_REQUEST['alone'] || $msg->HasFileData())
	{
		echo '<div class="msg-view' . ($showAvatars ? ' with-avatars' : '') . '">';
		show_message($msg, $rev);

		// Show message and revision history
		$logEntries = UnbLog::FindMessageModeration($msg, $rev);
		if (count($logEntries))
		{
			echo '<div class="message-separator"></div>';

			echo '<p>' . unbtrc('moderation history') . unbcol() . '<br />';
			foreach ($logEntries as $logEntry)
			{
				echo unbtrc('logevent:' . $logEntry['EventId']) . ' &nbsp;–&nbsp; ';
				UnbUi::RelativeTime($logEntry['Time']);
				if (isset($logEntry['UserName']))
				{
					echo ' ' . unbtr('by') . ' ' . htmlspecialchars($logEntry['UserName']);
				}
				echo '<br />';
			}
			echo '</p>';
		}
		echo '</div>';   // .msg-view
	}
	else
	{
		$prevTags = null;
		$prevSubject = null;

		$depId = $msg->GetRefDEP();
		if ($depId != $msg->GetMessageId())
		{
			$depMsg = UnbMessage::FindByMessageId($depId);
			$depRev = $depMsg->GetSearchRevision(UnbMessageRevisionContentLevel::Subject, true);
		}
		else
		{
			$depMsg =& $msg;
			$depRev =& $rev;
		}
		echo '<h1 style="margin-bottom: 0.3em;">' . htmlspecialchars($depRev->GetSubject()) . '</h1>';
		$prevSubject = $depRev->GetSubject();

		$tags = $depRev->GetTags(true);
		$prevTags = $tags;
		if ($tags)
		{
			$count = 0;
			foreach ($tags as $tag)
			{
				#if ($count++) echo ', ';
				echo '<span class="tag" title="' . htmlspecialchars($tag->GetDescription()) . '">';
				echo $tag->GetTagName();
				if (($tag->GetFlags() & UnbTagFlag::UserData) && ($tag->GetTagData() != null))
					echo ': ' . htmlspecialchars($tag->GetTagData());
				echo '</span> ';
			}
		}

		echo '<div style="margin: 1em 0; border-top: solid 1px #c0c0c0;"></div>';

		echo '<table style="clear: both; width: 100%; table-layout: fixed;" cellspacing="0">';
		echo '<tr>';
		echo '<td style="padding: 0; vertical-align: top; width: 17em; padding-right: 2em;">';

		if (!isset($rootMessageId))
		{
			// No message is specified as tree root.
			// Test whether the specified message is visible in the message tree.
			$depId2 = $msg->GetRefDEP(intval($messageThreadMaxLevel / 2));
			if ($depId2 != $depId)
			{
				// A limited DEP search has returned a different root message.
				// Use that as tree root to ensure that the specified message is well visible on this page.
				$rootMessageId = $depId2;
			}
		}

		echo unbtrc('messages in this discussion') . unbcol();
		echo '<div class="msg-tree">';
		UnbUi::MsgTree(
			$msg->GetMessageId(),
			$rootMessageId,
			true,
			$messageThreadMaxLevel,
			'<a href="view_message.php?%q' . (isset($rootMessageId) ? '&root=' . $rootMessageId : '') . '#message-%i">%s</a>',
			'view_message.php?messageid=%i&root=%r');
		echo '</div>';

		echo '</td>';
		echo '<td style="padding: 0; vertical-align: top;">';

		echo '<div class="msg-view' . ($showAvatars ? ' with-avatars' : '') . '">';

		$count = 0;
		$items = UnbUi::GetTreeMessages();
		foreach ($items as &$item)
		{
			// Only display messages selected in the tree
			if ($item['selected'])
			{
				// Print separator between messages
				if ($count++ > 0)
				{
					echo '<div class="message-separator"></div>';
				}

				// Check for multiple replies to the previous message, on the same level as this
				$replies = $item['replies'];
				if (count($replies) > 1)
				{
					echo '<div class="message notice">';
					echo '<div class="content" style="margin-bottom: 0.2em;">';
					echo '<img src="img/multiple-replies.png" alt="" /> ';
					echo unbtrc('multiple replies') . unbcol();
					echo '</div>';   // .content
					echo '</div>';   // .message.notice

					// List each reply message
					foreach ($replies as $reply)
					{
						$msg = $reply['msg'];
						$rev = $reply['rev'];

						echo '<div class="message notice">';
						if ($showAvatars)
						{
							echo '<div class="avatarcontainer">';
							echo '<img src="img/avatar_' . $msg->GetOwnerUser()->GetLoginName() . '.jpg" style="max-width: 50%; max-height: 1.2em;" />';
							echo '</div>';   // .avatarcontainer
						}

						echo '<div class="content">';
						// Show reply user name, with link if not currently selected for display
						if (!$reply['selected'])
							echo '<a href="view_message.php?messageid=' . $msg->GetMessageId() . (isset($rootMessageId) ? '&root=' . $rootMessageId : '') . '#message-' . $msg->GetMessageId() . '">';
						echo '<strong>' . htmlspecialchars($msg->GetOwnerUser()->GetDisplayName()) . '</strong>';
						if (!$reply['selected'])
							echo '</a>';

						// Show subject if available and differing
						if ($rev->GetContentLevel() >= UnbMessageRevisionContentLevel::Subject &&
							strlen($rev->GetSubject()) &&
							$rev->GetSubject() != $prevSubject &&
							$rev->GetSubject() != 'Re: ' . $prevSubject)
						{
							echo ' &nbsp; ' . unbqt(htmlspecialchars($rev->GetSubject()));
						}
						// Show time and total replies count in this subtree
						echo ' &nbsp;–&nbsp; ';
						UnbUi::RelativeTime($msg->GetCreatedTime());
						echo ' &nbsp;–&nbsp; ' . unbtr('n replies altogether', $reply['replies-count']);
						if ($reply['selected'])
							echo ' &nbsp; ' . unbpar(unbtr('displayed below'));

						// Show moderation flags overview
						$msgFlags = $msg->GetModerationFlags();
						$flags_array = array();
						if ($rev->GetModerationState() == UnbMessageRevisionModerationState::WaitingForApproval)
							$flags_array[] = unbtr('msg:unapproved');
						if ($msgFlags & UnbMessageModerationFlag::Locked || $rev->GetModerationState() == UnbMessageRevisionModerationState::Locked)
							$flags_array[] = unbtr('msg:locked');
						if ($msgFlags & UnbMessageModerationFlag::Hidden)
							$flags_array[] = unbtr('msg:hidden');
						if ($flags_array)
							echo ' &nbsp;–&nbsp; <span class="modwarning">' . unbcap(join(', ', $flags_array)) . '</span>';

						echo '</div>';   // .content
						echo '</div>';   // .message.notice
					}

					echo '<div class="message-separator"></div>';
				}

				$showMsg = $item['msg'];
				$showRev = $item['rev'];
				if ($rev->GetMessageId() == $showMsg->GetMessageId() &&
					$rev->GetMessageRevisionId() != $showRev->GetMessageRevisionId())
				{
					// Another revision of this message was explicitly specified, so display that instead of the default revision
					$showRev = $rev;
				}
				show_message($showMsg, $showRev);
			}
		}

		echo '</div>';   // .msg-view
		echo '</td>';
		echo '</tr>';
		echo '</table>';
	}
}
?>

<script>
function startHighlight(id)
{
	DOM.addClass(id, "highlight");
	setTimeout(function() { DOM.addClass(id, "highlight-fade"); }, 5000);
	// Clean up
	setTimeout(function() { DOM.removeClass(id, "highlight highlight-fade"); }, 7000);
}

function historyChanged()
{
	if (window.location.hash.match(/^#message-[0-9]+$/))
	{
		startHighlight(window.location.hash + "-div");
	}
}

historyChanged();

DOM.addHistoryChangedHandler(historyChanged);
</script>

<?php
require('_foot.inc.php');

// ---------- Page helper functions ----------

// Prints the normal reading view of a single message
//
function show_message($msg, $rev)
{
	global $markup, $prevTags, $prevSubject, $showAvatars;

	$isFile = $msg->HasFileData();

	// ----- Begin message block -----

	echo '<a id="message-' . $msg->GetMessageId() . '" style="position: relative; top: -2em;"></a>';
	echo '<div class="message" id="message-' . $msg->GetMessageId() . '-div">';

	if ($showAvatars)
	{
		echo '<div class="avatarcontainer">';
		echo '<img src="img/avatar_' . $msg->GetOwnerUser()->GetLoginName() . '.jpg" style="max-width: 100%; max-height: 5em;" />';
		echo '</div>';   // .avatarcontainer
	}

	// ----- Side actions -----

	echo '<div class="sideactions">';
	echo '<div>';
	if ($msg->GetOwnerUser()->IsAdministrator() && $msg->GetOwnerUser()->IsModerator())
		echo unbtrc('author is admin and mod') . '<br />';
	else if ($msg->GetOwnerUser()->IsAdministrator())
		echo unbtrc('author is admin') . '<br />';
	else if ($msg->GetOwnerUser()->IsModerator())
		echo unbtrc('author is mod') . '<br />';
	if (!$isFile &&
		$msg->GetSearchRevisionId() == $rev->GetMessageRevisionId() &&
		$rev->GetContentLevel() >= UnbMessageRevisionContentLevel::Content &&
		$msg->IsReplyAllowed())
		echo '<strong><a href="create_message.php?reprid=' . $rev->GetMessageRevisionId() . '">' . unbtrc('reply to this message') . '</a></strong><br />';
	// TODO: This is the wrong condition. Only the *last* revision may be altered, not the search revision
	if ($msg->GetSearchRevisionId() == $rev->GetMessageRevisionId() &&
		($msg->IsAlterAllowed() || UnbSession::IsAdministrator() || UnbSession::GetSessionUserId() == $msg->GetOwnerUserId()))
	{
		if (!$isFile)
			echo '<a href="create_message.php?altmid=' . $msg->GetMessageId() . '">' . unbtrc('alter this message') . '</a><br />';
		else
			echo '<a href="create_file.php?altmid=' . $msg->GetMessageId() . '">' . unbtrc('alter this file') . '</a><br />';
	}
	echo '<a href="view_user.php?userid=' . $msg->GetOwnerUserId() . '">' . unbtrc('view user profile') . '</a><br />';
	echo '</div>';

	if ($rev->GetCreatedTime() != $msg->GetCreatedTime())
	{
		echo '<div>';
		echo unbtrc('updated') . ' <a href="list_revisions.php?messageid=' . $msg->GetMessageId() . '">';
		UnbUi::RelativeTime($rev->GetCreatedTime());
		echo '</a>';
		echo '</div>';
	}
	else if ($msg->CountRevisions(true) > 1)
	{
		echo '<div>';
		#echo UnbLocale::OrdinalNumber($rev->GetRevisionNumber(), true) . ' revision updated ';
		#UnbUi::RelativeTime($rev->GetCreatedTime());
		echo '<a href="list_revisions.php?messageid=' . $msg->GetMessageId() . '">' . unbtrc('list all revisions') . '</a>';
		echo '</div>';
	}

	if ($msg->GetSearchRevisionId() == $rev->GetMessageRevisionId() &&
		$rev->GetContentLevel() >= UnbMessageRevisionContentLevel::Content &&
		UnbSession::IsLoggedIn() &&
		$msg->GetOwnerUserId() != UnbSession::GetSessionUserId())
	{
		echo '<div>';
		$myRating = $msg->GetMyRating();
		if (isset($myRating))
			$myRating = round($myRating / 5.0) * 5;
		echo unbtrc('rate') . unbcol() . ' ';
		UnbUi::MsgRateLink($msg, '<a class="rate-negative' . ($myRating == -10 ? ' selected' : '') . '" href="view_message.php?%s">&minus;&minus;</a>', -10);
		UnbUi::MsgRateLink($msg, '<a class="rate-negative' . ($myRating == -5 ? ' selected' : '') . '" href="view_message.php?%s">&nbsp;&minus;&nbsp;</a>', -5);
		UnbUi::MsgRateLink($msg, '<a class="rate-positive' . ($myRating == 5 ? ' selected' : '') . '" href="view_message.php?%s">&nbsp;+&nbsp;</a>', 5);
		UnbUi::MsgRateLink($msg, '<a class="rate-positive' . ($myRating == 10 ? ' selected' : '') . '" href="view_message.php?%s">++</a>', 10);
		if (isset($myRating))
			UnbUi::MsgClearRatingLink($msg, '&nbsp; <a href="view_message.php?%s">' . unbtr('clear rating') . '</a>');
		echo '</div>';
	}

	if (UnbSession::IsModerator())
	{
		echo '<div>';
		UnbUi::MsgLockRevLink($rev, '<a href="view_message.php?%s">' . unbtrc('lock this revision') . '</a><br />');
		UnbUi::MsgUnlockRevLink($rev, '<a href="view_message.php?%s">' . unbtrc('unlock this revision') . '</a><br />');
		echo '<a href="delete_message.php?messageid=' . $msg->GetMessageId() . '">' . unbtrc('delete this message') . '…</a><br />';
		echo '</div>';
	}
	echo '</div>';   // .sideactions

	// ----- Meta line above message -----

	echo '<div class="metaline">';
	echo '<span class="permalink">';
	echo '<a href="view_message.php?messageid=' . $msg->GetMessageId() . '#message-' . $msg->GetMessageId() . '" title="' . unbtrc('permalink') . '">#</a>';
	if (!$_REQUEST['alone'])
		echo ' <a href="view_message.php?messageid=' . $msg->GetMessageId() . '&amp;alone=1" title="' . unbtrc('view message alone') . '">&gt;</a>';
	echo '</span>';

	echo '<span class="author">';
	echo htmlspecialchars($msg->GetOwnerUser()->GetDisplayName());
	echo '</span>';

	$flags = $msg->GetModerationFlags();
	if ($rev->GetContentLevel() >= UnbMessageRevisionContentLevel::Subject)
	{
		if (strlen($rev->GetSubject()))
		{
			if ($rev->GetSubject() != $prevSubject &&
				$rev->GetSubject() != 'Re: ' . $prevSubject)
			{
				echo '<span class="subject">' . unbqt(htmlspecialchars($rev->GetSubject())) . '</span>';
			}
			$prevSubject = $rev->GetSubject();
		}
	}
	else
	{
		#echo '<span class="subject not-accessible">Subject not accessible</span>';
	}

	echo '<span class="date">';
	UnbUi::RelativeTime($msg->GetCreatedTime());
	echo '</span>';

	UnbUi::MsgRating($msg, 1, 1 /* TODO: Reset to 30 for release */, 30, '|msg-rating');

	$tags = $rev->GetTags(true);
	if (!isset($prevTags))
	{
		if ($tags)
		{
			$count = 0;
			echo '<br />';
			foreach ($tags as $tag)
			{
				#if ($count++) echo ', ';
				echo '<span class="tag" title="' . htmlspecialchars($tag->GetDescription()) . '">';
				echo $tag->GetTagName();
				if (($tag->GetFlags() & UnbTagFlag::UserData) && ($tag->GetTagData() != null))
					echo unbcol() . ' ' . htmlspecialchars($tag->GetTagData());
				echo '</span> ';
			}
		}
	}
	else
	{
		// Second or later revision: Show tag differences
		UnbTag::GetDifferences($prevTags, $tags, $addedTags, $removedTags, $changedTags);
		$str = '';
		foreach ($addedTags as $tag)
		{
			$str .= '<span class="tag added">+';
			$str .= $tag->GetTagName();
			if (($tag->GetFlags() & UnbTagFlag::UserData) && ($tag->GetTagData() != null))
				$str .= unbcol() . ' ' . htmlspecialchars($tag->GetTagData());
			$str .= '</span> ';
		}
		foreach ($changedTags as $tag)
		{
			$str .= '<span class="tag changed">*';
			$str .= $tag->GetTagName();
			if (($tag->GetFlags() & UnbTagFlag::UserData) && ($tag->GetTagData() != null))
				$str .= unbcol() . ' ' . htmlspecialchars($tag->GetTagData());
			$str .= '</span> ';
		}
		foreach ($removedTags as $tag)
		{
			$str .= '<span class="tag removed">&minus;';
			$str .= $tag->GetTagName();
			if (($tag->GetFlags() & UnbTagFlag::UserData) && ($tag->GetTagData() != null))
				$str .= unbcol() . ' ' . htmlspecialchars($tag->GetTagData());
			$str .= '</span> ';
		}
		if (strlen($str))
		{
			echo '<br />' . $str;
		}
	}
	$prevTags = $tags;

	// ----- Prepare file attachments list -----

	$fileIds = $rev->GetFileIds();

	$unapproved = $rev->GetModerationState() == UnbMessageRevisionModerationState::WaitingForApproval;
	$isLocked = $flags & UnbMessageModerationFlag::Locked || $flags & UnbMessageModerationFlag::Hidden;
	if (UnbSession::IsModerator() || $msg->GetOwnerUserId() == UnbSession::GetSessionUserId())
		$isLocked = false;

	$unapprovedFiles = false;
	$attachStr = '';
	if (!$isLocked && count($fileIds))
	{
		$attachStr .= '<div class="message-attachments">';
		if ($unapproved && !UnbSession::IsModerator() && $rev->GetAuthorUserId() != UnbSession::GetSessionUserId())
		{
			$attachStr .= unbtrc('n files become visible upon approval', count($fileIds));
		}
		else
		{
			foreach ($fileIds as $fileId)
			{
				$fileMsg = UnbMessage::FindByMessageId($fileId);
				$fileRev = $fileMsg->GetSearchOrLatestRevision(UnbMessageRevisionContentLevel::Content, true);

				$size = $fileRev->GetFileDataLength();
				if ($fileRev->GetContentLevel() >= UnbMessageRevisionContentLevel::Subject)
				{
					$fileName = htmlspecialchars($fileRev->GetSubject());
					if (!strlen($fileName))
						$fileName = '<em>' . unbtrc('unnamed file') . '</em>';
				}
				else
				{
					$fileName = '<em>' . unbtrc('file name not accessible') . '</em>';
				}
				if ($fileRev->GetContentLevel() >= UnbMessageRevisionContentLevel::Content)
				{
					$contentType = $fileRev->GetContentType();
				}
				else
				{
					$contentType = '';
				}

				$attachStr .= '<div class="item">';
				$attachStr .= UnbUi::ContentTypeIcon($contentType, true);
				$attachStr .= '<a href="get_data.php?messageid=' . $fileId . '">' . $fileName . '</a>';
				$attachStr .= '<span class="desc"> &nbsp;–&nbsp; ';
				$attachStr .= htmlspecialchars(UnbLocale::Capital(UnbLocale::ContentType($contentType))) . ', ';
				$attachStr .= UnbLocale::DataSize($size, true) . ', ';
				$attachStr .= '<a href="view_message.php?messageid=' . $fileId . '">' . unbtrc('view details') . '</a></span>';

				$msgFlags = $fileMsg->GetModerationFlags();
				$flags_array = array();
				if ($fileRev->GetModerationState() == UnbMessageRevisionModerationState::WaitingForApproval)
				{
					$flags_array[] = unbtr('msg:unapproved');
					$unapprovedFiles = true;
				}
				if ($msgFlags & UnbMessageModerationFlag::Locked || $fileRev->GetModerationState() == UnbMessageRevisionModerationState::Locked)
					$flags_array[] = unbtr('msg:locked');
				if ($msgFlags & UnbMessageModerationFlag::Hidden)
					$flags_array[] = unbtr('msg:hidden');
				if ($flags_array)
					echo ' &nbsp;–&nbsp; <span class="modwarning">' . unbcap(join(', ', $flags_array)) . '</span>';

				$attachStr .= '</div>';   // .item
			}
		}
		$attachStr .= '</div>';   // .message-attachments
	}

	// ----- Informational and moderative messages -----

	if ($rev->IsDraft())
	{
		echo '<div class="modwarning">' . unbtrc('this is a draft');
		if (UnbSession::GetSessionUserId() == $rev->GetAuthorUserId())
			echo ' <a href="create_message.php?drarid=' . $rev->GetMessageRevisionId() . '">' . unbtrc('continue editing') . '</a>';
		echo '</div>';
	}
	if ($msg->GetSearchRevisionId() > $rev->GetMessageRevisionId())
	{
		echo '<div class="modwarning">' . unbtrc('not the latest available revision of this message') . ' <a href="view_message.php?messageid=' . $msg->GetMessageId() . '#message-' . $msg->GetMessageId() . '">' . unbtrc('view latest revision') . '</a></div>';
	}
	if ($flags & UnbMessageModerationFlag::Hidden)
	{
		echo '<div class="modwarning">' . unbtrc('msg:mod.hidden long') . '</div>';
	}
	if ($flags & UnbMessageModerationFlag::Locked)
	{
		echo '<div class="modwarning">' . unbtrc('msg:mod.locked long') . '</div>';
	}
	if ($unapproved)
	{
		echo '<div class="modwarning">' . unbtrc('msg:mod.revision waiting for approval long');
		if (UnbSession::IsModerator())
			UnbUi::MsgApproveRevLink($rev, ' <a href="view_message.php?%s">' . unbtrc('approve now' . (count($fileIds) ? ' with all files' : '')) . '</a><br />');
		echo '</div>';
	}
	else if ($rev->GetModerationState() == UnbMessageRevisionModerationState::Locked)
	{
		echo '<div class="modwarning">' . unbtrc('msg:mod.revision locked long') . '</div>';
	}
	else if ($unapprovedFiles)
	{
		echo '<div class="modwarning">' . unbtrc('msg:mod.files waiting for approval long') . '</div>';
	}

	$readAccessKeys = $msg->GetEffectiveReadAccessKeys();
	if (isset($readAccessKeys))
	{
		$userNames = UnbUi::GetUserList($readAccessKeys);
		echo '<div class="modwarning">' . unbtrc('read access for') . unbcol() . ' ' . str_replace("\n", ', ', $userNames) . '</div>';
	}
	echo '</div>';   // .metaline

	// ----- Message content -----

	echo '<div class="content">';
	if ($rev->GetContentLevel() >= UnbMessageRevisionContentLevel::HtmlContent)
	{
		if ($msg->HasFileData() && preg_match('~^image/~', $rev->GetContentType()))
		{
			// Display the attached image file
			echo '<p><img src="get_data.php?messagerevisionid=' . $rev->GetMessageRevisionId() . '" alt="' . unbtrc('attached image') . '" style="max-width: 100%;" /></p>';
		}

		$html = $rev->GetHtmlContent();
		#$html = null;   // DEBUG: Don't use cached HTML
		if (!isset($html))
		{
			if ($rev->GetContent() !== null)
			{
				$html = $markup->Convert($rev->GetContent());
				try
				{
					$rev->SetHtmlContent($html);
				}
				catch (Exception $ex) { }   // Ignore errors here
			}
			else
			{
				$html = '';   // Empty contents should always be empty HTML
			}
		}
		$html = $markup->PostProcess($html);
		if (strlen($html) > 0)
		{
			echo '<div class="markup">' . $html . '</div>';
		}
		else
		{
			if (!$msg->HasFileData())
				echo '<p><em>' . unbtrc('message is empty') . '</em></p>';
			else
				echo '<p><em>' . unbtrc('file has no description') . '</em></p>';
		}
	}
	else
	{
		echo '<div class="modwarning">' . unbtrc('content not accessible') . '</div>';
	}

	// ----- File information -----

	if ($msg->HasFileData())
	{
		try
		{
			$size = $rev->GetFileDataLength();
			if ($rev->GetContentLevel() >= UnbMessageRevisionContentLevel::Content)
			{
				$contentType = $rev->GetContentType();
			}
			else
			{
				$contentType = '';
			}

			echo '<p>';
			echo '<strong><a href="get_data.php?messagerevisionid=' . $rev->GetMessageRevisionId() . '">' . unbtrc('download the file') . '</a></strong><br />';
			echo unbtrc('file type') . unbcol() . ' ';
			UnbUi::ContentTypeIcon($contentType);
			echo htmlspecialchars(UnbLocale::Capital(UnbLocale::ContentType($contentType))) . '<br />';
			echo unbtrc('file size') . unbcol() . ' ' . UnbLocale::DataSize($size, true) . '<br />';
			echo '</p>';
		}
		catch (Exception $ex)
		{
			echo '<p style="color: red;">There is a problem with the attached data: ' . $ex->getMessage() . '</p>';
		}
	}

	// ----- Print file attachments list -----

	echo $attachStr;

	// ----- "Outro" -----

	// TODO: Polish this a bit, then translate it
	$flags_str = '';
	if ($flags & UnbMessageModerationFlag::EnforceApproval) $flags_str .= ' EnforceApproval';
	if ($flags_str)
		echo '<p><small>Flags: ' . trim($flags_str) . '</small></p>';
	echo '</div>';   // .content

	echo '<div style="clear: both;"></div>';
	echo '</div>';   // .message
}
?>
