SET sql_mode = 'ANSI';
USE "unb2test";

-- Create keylist index table
ALTER TABLE "keylist" RENAME "keylist_keys";
CREATE TABLE "keylist" ("KeylistId" INTEGER UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY);
INSERT INTO "keylist" SELECT DISTINCT "KeylistId" FROM "keylist_keys";

-- Generate new ordered message revision IDs
CREATE TABLE "message_revision2" LIKE "message_revision";
ALTER TABLE "message_revision2"
	ADD "MessageRevisionId" INTEGER UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY;
INSERT INTO "message_revision2" SELECT * FROM "message_revision" ORDER BY "MessageId", "RevisionNumber";
ALTER TABLE "message_revision" RENAME "message_revision1";
ALTER TABLE "message_revision2" RENAME "message_revision";

-- Update the messages' search revisions from revision numbers to IDs
ALTER TABLE "message"
	CHANGE "SearchRevision" "SearchRevisionId" INTEGER UNSIGNED;
UPDATE "message" m
	SET "SearchRevisionId" =
		(SELECT "MessageRevisionId"
			FROM "message_revision" mr
			WHERE mr."MessageId" = m."MessageId"
				AND mr."RevisionNumber" = m."SearchRevisionId")
	WHERE "SearchRevisionId" IS NOT NULL;

-- Update message revision tags to the new revision ID key
ALTER TABLE "message_revision_tag"
	ADD "MessageRevisionId" INTEGER UNSIGNED NOT NULL;
UPDATE "message_revision_tag" mrt
	SET "MessageRevisionId" =
		(SELECT "MessageRevisionId"
			FROM "message_revision" mr
			WHERE mr."MessageId" = mrt."MessageId"
				AND mr."RevisionNumber" = mrt."RevisionNumber");

-- Use the following shell commands to rename the attachment files
SELECT 'mv ' || CAST("MessageId" AS CHAR) || '_' || CAST("RevisionNumber" AS CHAR) || ' '
	|| CAST("MessageId" AS CHAR) || '_' || CAST("MessageRevisionId" AS CHAR) AS "command"
FROM "message_revision"
WHERE "MessageRevisionId" <> "RevisionNumber"
ORDER BY "MessageId", "RevisionNumber" DESC;

-- Update message attachment indicators
ALTER TABLE "message" ADD "FileType" TINYINT UNSIGNED;
UPDATE "message" SET "FileType" =
    (SELECT CASE "HasData" WHEN TRUE THEN 1 ELSE 0 END
    FROM "message_revision"
    WHERE "MessageRevisionId" = "SearchRevisionId");

-- Add message revisions' embedded files table
CREATE TABLE "message_revision_file" (
	"MessageRevisionId" INTEGER UNSIGNED NOT NULL,
	"MessageId" INTEGER UNSIGNED NOT NULL,
	"Relation" VARCHAR(255),
	PRIMARY KEY ("MessageRevisionId", "MessageId"))
ENGINE 'InnoDB' CHARACTER SET 'utf8' COLLATE 'utf8_bin';

-- Add attachment data content type column
ALTER TABLE "message_revision"
	ADD "ContentType" VARCHAR(255);

-- Add draft columns
ALTER TABLE "message"
	ADD "Draft" BOOLEAN NOT NULL DEFAULT FALSE;

ALTER TABLE "message_revision"
	ADD "Draft" BOOLEAN NOT NULL DEFAULT FALSE;

-- Now export the data, recreate the tables and import the data.
-- (The table schema cannot be altered online, referential integrity is defending it.)
