<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// Database import script

// Encoding: UTF-8 (without BOM) (auto-detect: °°°°°)
header('Content-Type: text/html; charset=utf-8');

// USE WITH CARE!
//
// For security reasons, all configuration must be done statically
// in this source file. There is no web interface for this script.
//
// Be sure that the tables of this UNB setup are empty!
//
// Configuration should only be required between the lines below:

// ————————————————————————————————————————————————————————————————
// Comment the following line ONLY to enable the script.
// Be sure to undo it as soon as you're done.
#exit();

// Enter the file name to dump the database into:
$fileName = 'unb_dump.txt';

// Include the complete UNB library
require_once('unb_lib/unb_all.lib.php');
// Settings file name is relative to the UNB library source files
UnbUi::Init('../unb_config.inc.php');
// ————————————————————————————————————————————————————————————————

function decode($str)
{
	if ($str === '%NULL')
	{
		return null;
	}
	$str = preg_replace('_%([0-9A-Fa-f]{2})_e', 'chr(hexdec("$1"))', $str);   // like urldecode
	return $str;
}

function insert($table, $values)
{
	$columns = '';
	$q = '';
	for ($i = 2; $i < func_num_args(); $i++)
	{
		$columns .= ($columns ? ', ' : '') .
			'"' . func_get_arg($i) . '"';
		$q .= ($q ? ', ' : '') .
			'?';
	}

	if ($table == 'user')
	{
		$columns .= ', "DisplayNamePrep"';
		$q .= ', ?';
		$values[] = UnbUser::PrepDisplayName($values[3]);
	}

	UnbDatabase::GetInstance()->Execute('INSERT INTO "#' . $table . '" (' . $columns . ') VALUES (' . $q . ')', $values);
}

function read($f)
{
	$line = fgets($f);
	if ($line === false)
		return false;
	$fields = explode(';', trim($line));
	$table = $fields[0];
	$fields = array_slice($fields, 1);
	foreach ($fields as &$value)
	{
		$value = decode($value);
	}

	switch ($table)
	{
		case 'keylist':
			insert($table, $fields, 'KeylistId');
			break;
		case 'keylist_keys':
			insert($table, $fields, 'KeylistId', 'UserId');
			break;
		case 'user':
			insert($table, $fields, 'UserId', 'CreatedTime', 'LastLoginTime', /*[3]*/ 'DisplayName', 'AdditionalKeylistId', 'LoginName', 'LoginPassword');
			break;
		case 'user_vcard':
			insert($table, $fields, 'UserId', 'Type', 'Data', 'Visibility');
			break;
		case 'user_config':
			insert($table, $fields, 'UserId', 'Name', 'Data');
			break;
		case 'message':
			insert($table, $fields, 'MessageId', 'CreatedTime', 'Owner', 'PrimaryReference', 'EntryPoint', 'PageName', 'Draft', 'ReadAccessKeylistId', 'AlterAccessKeylistId', 'ReplyAccessKeylistId', 'ModerationFlags', 'FileType', 'SearchOptimisedContent', 'SearchOptimisedSubject', 'SearchRevisionId', 'LastBeginEdit');
			break;
		case 'message_reference':
			insert($table, $fields, 'MessageId', 'Reference');
			break;
		case 'message_revision':
			insert($table, $fields, 'MessageRevisionId', 'MessageId', 'CreatedTime', 'Author', 'Subject', 'Content', /*'HtmlContent',*/ 'Summary', 'ModerationState', 'Draft', 'ContentType', 'Data');
			break;
		case 'tag':
			insert($table, $fields, 'TagId', 'TagName', 'ParentTag', 'Meaning', 'Description', 'UseAccessKeylistId', 'ReadAccessKeylistId', 'Flags');
			break;
		case 'message_revision_tag':
			insert($table, $fields, 'MessageRevisionId', 'TagId', 'TagData');
			break;
		case 'message_revision_file':
			insert($table, $fields, 'MessageRevisionId', 'MessageId', 'Relation');
			break;
		case 'message_rating':
			insert($table, $fields, 'UserId', 'MessageId', 'Rating', 'Weight');
			break;
	}
	return true;
}

// Do the job...
$f = fopen($fileName, 'r');

UnbDatabase::GetInstance()->StartTransaction(true);

// TODO: Required for MySQL, must be omitted for SQLite
UnbDatabase::GetInstance()->Execute('SET FOREIGN_KEY_CHECKS = 0');

while (read($f));

UnbDatabase::GetInstance()->Execute('SET FOREIGN_KEY_CHECKS = 1');

UnbDatabase::GetInstance()->Commit();

fclose($f);

UnbSession::SetSessionUser(null);
echo 'Import complete.';
?>