<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UNB locale en-GB class.

// Encoding: UTF-8 (without BOM) (auto-detect: °°°°°)

if (!class_exists('UnbLocale', false)) require_once dirname(__FILE__) . '/../locale.class.php';

// en-GB locale class.
//
class UnbLocale_en_GB extends UnbLocale
{
	// Locale code.
	protected $code = 'en-GB';

	// Locale importance for its language, used to sort known locales when only a language is requested.
	// This should be roughly the number of people affected by the locale, divided by 1000.
	protected $importance = 60000;   // Source: http://en.wikipedia.org/wiki/English_language#Countries_in_order_of_total_speakers

	// Locale code that the locale depends on for dictionary translations.
	protected $dependsOn = 'en-US';

	// Starting quotation mark.
	protected $quoteStart = '“';

	// Ending quotation mark.
	protected $quoteEnd = '”';

	// Colon.
	protected $colon = ':';

	// The text dictionary.
	protected $dict = array(
		);

	protected function FormatNumber($number, $decimals, $html)
	{
		return number_format($number, $decimals, '.', ',');
	}

	protected function FormatOrdinalNumber($number, $html)
	{
		if ($number <= 0) return $num;
		if ($number % 10 == 1) return $number . ($html ? '<sup>' : '') . 'st' . ($html ? '</sup>' : '');
		if ($number % 10 == 2) return $number . ($html ? '<sup>' : '') . 'nd' . ($html ? '</sup>' : '');
		if ($number % 10 == 3) return $number . ($html ? '<sup>' : '') . 'rd' . ($html ? '</sup>' : '');
		return $number . ($html ? '<sup>' : '') . 'th' . ($html ? '</sup>' : '');
	}

	protected function FormatTime($datetime, $details, $html)
	{
		$sp = "\xC2\xA0";   // "\xC2\xA0" = utf8_encode("\xA0")

		$str = '';

		if ($details & UnbTimeFormat::Weekday)   // Mon
		{
			$str .= (strlen($str) ? $sp : '') .
				UnbLocale::Translate('weekday:' . $datetime->format('N')) .
				($details & ~UnbTimeFormat::Weekday ? ',' : '');
		}
		else if ($details & UnbTimeFormat::WeekdayLong)   // Monday
		{
			$str .= (strlen($str) ? $sp : '') .
				UnbLocale::Translate('weekdaylong:' . $datetime->format('N')) .
				($details & ~UnbTimeFormat::WeekdayLong ? ',' : '');
		}

		if ($details & UnbTimeFormat::Date)   // 01/01/2000
		{
			$str .= (strlen($str) ? $sp : '') .
				$datetime->format('d/m/Y');
			$haveDate = true;
		}
		else if ($details & UnbTimeFormat::DateLong)   // 1st January 2000
		{
			$str .= (strlen($str) ? $sp : '') .
				$this->FormatOrdinalNumber($datetime->format('j'), $html) . $sp .
				UnbLocale::Translate('month:' . $datetime->format('n')) . $sp .
				$datetime->format('Y');
			$haveDate = true;
		}

		if ($details & UnbTimeFormat::TimeSec)   // 12:15:00 pm
		{
			$str .= (strlen($str) ? ($haveDate && $html ? $sp . $sp : $sp) : '') .
				$datetime->format('g:i:s a');
		}
		else if ($details & UnbTimeFormat::Time)   // 12:15 pm
		{
			$str .= (strlen($str) ? ($haveDate && $html ? $sp . $sp : $sp) : '') .
				$datetime->format('g:i a');
		}

		if ($details & UnbTimeFormat::TimeZone)   // UTC
		{
			$str .= (strlen($str) ? $sp : '') .
				$datetime->format('T');
		}

		return $str;
	}

	protected function FormatRelativeTime($diff, $context, $html)
	{
		switch ($context)
		{
			case UnbTimeContext::Point:
				if ($diff->invert)   // 2 years ago
					$grammar = 1;
				else   // in 2 years
					$grammar = 1;
				break;

			case UnbTimeContext::Span:
				if ($diff->invert)   // for 2 years
					$grammar = 1;
				else   // for 2 years
					$grammar = 1;
				break;

			case UnbTimeContext::Before:
				if ($diff->invert)   // 2 years before
					$grammar = 1;
				else   // 2 years later
					$grammar = 1;
				break;
		}

		$parts = array();
		if ($diff->y && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':years', $diff->y));
		if ($diff->m && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':months', $diff->m));
		if ($diff->d && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':days', $diff->d));
		if ($diff->h && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':hours', $diff->h));
		if ($diff->i && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':minutes', $diff->i));
		if ($diff->s && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':seconds', $diff->s));

		$str = '';
		foreach ($parts as $part)
		{
			$str .= (strlen($str) ? UnbLocale::Translate('time:combine') : '') .
				$part;
		}

		switch ($context)
		{
			case UnbTimeContext::Point:
				if ($diff->invert)   // 2 years ago
					$str = self::Translate('time:pattern point past', 'time', $str);
				else   // in 2 years
					$str = self::Translate('time:pattern point future', 'time', $str);
				break;

			case UnbTimeContext::Span:
				if ($diff->invert)   // for 2 years
					$str = self::Translate('time:pattern span past', 'time', $str);
				else   // for 2 years
					$str = self::Translate('time:pattern span future', 'time', $str);
				break;

			case UnbTimeContext::Before:
				if ($diff->invert)   // 2 years before
					$str = self::Translate('time:pattern before past', 'time', $str);
				else   // 2 years later
					$str = self::Translate('time:pattern before future', 'time', $str);
				break;
		}
		return $str;
	}

}

// Automatically register this locale
UnbLocale::Register(new UnbLocale_en_GB());

?>
