<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UNB locale fr-FR class.

// Encoding: UTF-8 (without BOM) (auto-detect: °°°°°)

if (!class_exists('UnbLocale', false)) require_once dirname(__FILE__) . '/../locale.class.php';

// fr-FR locale class.
//
class UnbLocale_fr_FR extends UnbLocale
{
	// Locale code.
	protected $code = 'fr-FR';

	// Locale importance for its language, used to sort known locales when only a language is requested.
	// This should be roughly the number of people affected by the locale, divided by 1000.
	protected $importance = -1;   // TODO: unknown

	// Starting quotation mark.
	protected $quoteStart = '« ';

	// Ending quotation mark.
	protected $quoteEnd = ' »';

	// Colon.
	protected $colon = ' :';

	// The text dictionary.
	protected $dict = array(
		'time:none' => 'pas de temps',
		'time:never' => 'jamais',
		'time:now' => 'maintenant',
		'time:combine' => ' et ',
		'time:pattern point past' => 'il y a {time}',
		'time:pattern point future' => 'dans {time}',
		'time:pattern span past' => 'depuis {time}',
		'time:pattern span future' => 'pour {time}',
		'time:pattern before past' => '{time} avant',
		'time:pattern before future' => '{time} après',
		'time.1:years' => '{#} ans',
		'time.1:years#1' => 'un an',
		'time.1:months' => '{#} mois',
		'time.1:months#1' => 'un mois',
		'time.1:days' => '{#} jours',
		'time.1:days#1' => 'un jour',
		'time.1:hours' => '{#} heures',
		'time.1:hours#1' => 'une heure',
		'time.1:minutes' => '{#} minutes',
		'time.1:minutes#1' => 'une minute',
		'time.1:seconds' => '{#} secondes',
		'time.1:seconds#1' => 'une seconde',
		'month:1' => 'janvier',
		'month:2' => 'février',
		'month:3' => 'mars',
		'month:4' => 'avril',
		'month:5' => 'mai',
		'month:6' => 'juin',
		'month:7' => 'juillet',
		'month:8' => 'août',
		'month:9' => 'septembre',
		'month:10' => 'octobre',
		'month:11' => 'novembre',
		'month:12' => 'décembre',
		'weekday:1' => 'lu',
		'weekday:2' => 'ma',
		'weekday:3' => 'me',
		'weekday:4' => 'je',
		'weekday:5' => 've',
		'weekday:6' => 'sa',
		'weekday:7' => 'di',
		'weekdaylong:1' => 'lundi',
		'weekdaylong:2' => 'mardi',
		'weekdaylong:3' => 'mercredi',
		'weekdaylong:4' => 'jeudi',
		'weekdaylong:5' => 'vendredi',
		'weekdaylong:6' => 'samedi',
		'weekdaylong:7' => 'dimanche',
		);

	protected function FormatNumber($number, $decimals, $html)
	{
		return number_format($number, $decimals, ',', '.');
	}

	protected function FormatOrdinalNumber($number, $html)
	{
		return $number . '.';

		// TODO
		if ($number <= 0) return $num;
		if ($number % 10 == 1) return $number . ($html ? '<sup>' : '') . 'st' . ($html ? '</sup>' : '');
		if ($number % 10 == 2) return $number . ($html ? '<sup>' : '') . 'nd' . ($html ? '</sup>' : '');
		if ($number % 10 == 3) return $number . ($html ? '<sup>' : '') . 'rd' . ($html ? '</sup>' : '');
		return $number . ($html ? '<sup>' : '') . 'th' . ($html ? '</sup>' : '');
	}

	protected function FormatTime($datetime, $details, $html)
	{
		$sp = "\xC2\xA0";   // "\xC2\xA0" = utf8_encode("\xA0")

		$str = '';

		if ($details & UnbTimeFormat::Weekday)   // Lun
		{
			$str .= (strlen($str) ? $sp : '') .
				UnbLocale::Translate('weekday:' . $datetime->format('N'));
		}
		else if ($details & UnbTimeFormat::WeekdayLong)   // Lundi
		{
			$str .= (strlen($str) ? $sp : '') .
				UnbLocale::Translate('weekdaylong:' . $datetime->format('N'));
		}

		if ($details & UnbTimeFormat::Date)   // 01/01/2000
		{
			$str .= (strlen($str) ? $sp : '') .
				$datetime->format('d/m/Y');
			$haveDate = true;
		}
		else if ($details & UnbTimeFormat::DateLong)   // 1 janvier 2000
		{
			$str .= (strlen($str) ? $sp : '') .
				$datetime->format('j') . $sp .
				UnbLocale::Translate('month:' . $datetime->format('n')) . $sp .
				$datetime->format('Y');
			$haveDate = true;
		}

		if ($details & UnbTimeFormat::TimeSec)   // 12:15:00
		{
			$str .= (strlen($str) ? ($haveDate && $html ? $sp . $sp : $sp) : '') .
				$datetime->format('H:i:s');
		}
		else if ($details & UnbTimeFormat::Time)   // 12:15
		{
			$str .= (strlen($str) ? ($haveDate && $html ? $sp . $sp : $sp) : '') .
				$datetime->format('H:i');
		}

		if ($details & UnbTimeFormat::TimeZone)   // UTC
		{
			$str .= (strlen($str) ? $sp : '') .
				$datetime->format('T');
		}

		return $str;
	}

	protected function FormatRelativeTime($diff, $context, $html)
	{
		// TODO

		switch ($context)
		{
			case UnbTimeContext::Point:
				if ($diff->invert)   // 2 years ago
					$grammar = 1;
				else   // in 2 years
					$grammar = 1;
				break;

			case UnbTimeContext::Span:
				if ($diff->invert)   // for 2 years
					$grammar = 1;
				else   // for 2 years
					$grammar = 1;
				break;

			case UnbTimeContext::Before:
				if ($diff->invert)   // 2 years before
					$grammar = 1;
				else   // 2 years later
					$grammar = 1;
				break;
		}

		$parts = array();
		if ($diff->y && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':years', $diff->y));
		if ($diff->m && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':months', $diff->m));
		if ($diff->d && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':days', $diff->d));
		if ($diff->h && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':hours', $diff->h));
		if ($diff->i && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':minutes', $diff->i));
		if ($diff->s && count($parts) < 2)
			$parts[] = UnbLocale::NoBreak(UnbLocale::Translate('time.' . $grammar . ':seconds', $diff->s));

		$str = '';
		foreach ($parts as $part)
		{
			$str .= (strlen($str) ? UnbLocale::Translate('time:combine') : '') .
				$part;
		}

		switch ($context)
		{
			case UnbTimeContext::Point:
				if ($diff->invert)   // 2 years ago
					$str = self::Translate('time:pattern point past', 'time', $str);
				else   // in 2 years
					$str = self::Translate('time:pattern point future', 'time', $str);
				break;

			case UnbTimeContext::Span:
				if ($diff->invert)   // for 2 years
					$str = self::Translate('time:pattern span past', 'time', $str);
				else   // for 2 years
					$str = self::Translate('time:pattern span future', 'time', $str);
				break;

			case UnbTimeContext::Before:
				if ($diff->invert)   // 2 years before
					$str = self::Translate('time:pattern before past', 'time', $str);
				else   // 2 years later
					$str = self::Translate('time:pattern before future', 'time', $str);
				break;
		}
		return $str;
	}

}

// Automatically register this locale
UnbLocale::Register(new UnbLocale_fr_FR());

?>
