<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbUi class.

// Encoding: UTF-8 (without BOM) (auto-detect: °°°°°)

if (!class_exists('UnbDatabase', false)) require_once dirname(__FILE__) . '/database.class.php';
if (!class_exists('UnbSession', false)) require_once dirname(__FILE__) . '/session.class.php';
if (!class_exists('UnbUser', false)) require_once dirname(__FILE__) . '/user.class.php';
if (!class_exists('UnbMessage', false)) require_once dirname(__FILE__) . '/message.class.php';
if (!class_exists('UnbMessageRevision', false)) require_once dirname(__FILE__) . '/messagerevision.class.php';
if (!class_exists('UnbTag', false)) require_once dirname(__FILE__) . '/tag.class.php';
if (!class_exists('UnbSecurityException', false)) require_once dirname(__FILE__) . '/unb.class.php';
if (!class_exists('UnbSearch', false)) require_once dirname(__FILE__) . '/search.class.php';
if (!class_exists('UnbFormHandler', false)) require_once dirname(__FILE__) . '/formhandler.class.php';
if (!class_exists('UnbEnvironment', false)) require_once dirname(__FILE__) . '/environment.class.php';
if (!class_exists('UnbLocale', false)) require_once dirname(__FILE__) . '/locale.class.php';

// Contains methods to print out HTML controls that form the UI of UNB components, using the
// current UNB locale.
//
class UnbUi
{
	// Do automatic redirects to other pages. Disable this to debug form processing things.
	const DoRedirects = true;

	public static $data = array(
		'img_url' => 'img/',   // TODO: Consider project_back_path from UnbEnvironment
		'contenttype_img_url' => 'img/contenttype/',   // TODO: Consider project_back_path from UnbEnvironment
		'plaintype_help_url' => 'plaintype.html',   // TODO: Create the page and set its link, consider translations (translate the link URL)
		'plaintype_help_width' => 600,
		'plaintype_help_height' => 500,
		);

	private static $formHandler = null;
	private static $formFields = null;

	// Indicates whether the HTML block for message reporting shall be written.
	private static $useMessageBlock = false;
	// Indicates whether an HTML block for message reporting is currently opened.
	private static $messageBlockOpen = false;

	// Displayed messages by MsgTree
	private static $treeMessages = null;

	// Private constructor, does nothing but preventing an unwanted instantiation of this class.
	//
	private function __construct()
	{
	}

	// Ensures that the form has been handled.
	//
	private static function EnsureHandleForm()
	{
		if (!isset(self::$formHandler))
		{
			self::$formHandler = new UnbFormHandler();
			self::$formFields = self::$formHandler->HandleForm();
		}
	}

	// Gets the form handler instance.
	//
	// Returns (UnbFormHandler) Form handler instance
	//
	public static function GetFormHandler()
	{
		self::EnsureHandleForm();
		return self::$formHandler;
	}

	// Gets form handler return fields.
	//
	// action = (string) Action name to get fields for, or {{null}} to get all fields
	// field = (string) Field name to get, or {{null}} to get all fields
	//
	// Returns (array/any) Form field(s)
	//
	public static function GetFormHandlerFields($action = null, $field = null)
	{
		self::EnsureHandleForm();
		if (isset($action))
		{
			if (isset($field))
			{
				return self::$formFields['actions'][$action][$field];
			}
			else
			{
				return self::$formFields['actions'][$action];
			}
		}
		else
		{
			return self::$formFields;
		}
	}

	// Gets a value indicating whether the form handling has been successfully completed.
	//
	// Returns (bool)
	//
	public static function IsFormSuccess()
	{
		self::EnsureHandleForm();
		return isset(self::$formFields['allsuccess']);
	}

	// Gets a value indicating whether an error has occured during form handling.
	//
	// Returns (bool)
	//
	public static function IsFormError()
	{
		self::EnsureHandleForm();
		return isset(self::$formFields['erroraction']);
	}

	// Gets the form handler error message, if an error has occured.
	//
	// Returns (string) Error message or {{null}} if no error has occured
	//
	public static function GetFormErrorMessage()
	{
		self::EnsureHandleForm();
		$errorAction = self::$formFields['erroraction'];
		return self::$formFields['actions'][$errorAction]['error'];
	}

	// Formats and prints all form handler fields as an XML document.
	//
	// This function also sends the Content-Type header with text/xml type. You should not generate
	// any further output after this function call.
	//
	public static function FormHandlerFieldsXML()
	{
		$fields = self::GetFormHandlerFields();

		header('Content-Type: text/xml');
		self::FormHandlerFieldsXML_rec('fields', $fields, 0);
	}

	// Recursion helper function for FormHandlerFieldsXML.
	//
	private static function FormHandlerFieldsXML_rec($parentKey, $data, $level)
	{
		$pad = str_repeat("\t", $level);

		// Element names cannot contain digits only in XML
		if (is_numeric($parentKey))
			$parentKey = 'item';

		if (is_array($data))
		{
			if (count($data))
			{
				$isAssoc = false;
				foreach (array_keys($data) as $aKey)
					if (!is_numeric($aKey))
						$isAssoc = true;
				#$isAssoc = true;

				if ($isAssoc)
					echo $pad . '<' . $parentKey . '>' . "\n";

				foreach ($data as $key => $value)
				{
					if ($isAssoc)
						self::FormHandlerFieldsXML_rec($key, $value, $level + 1);
					else
						self::FormHandlerFieldsXML_rec($parentKey, $value, $level);
				}

				if ($isAssoc)
					echo $pad . '</' . $parentKey . '>' . "\n";
			}
		}
		else if (is_bool($data))
		{
			if ($data)
				echo $pad . '<' . $parentKey . '/>' . "\n";
		}
		else if (is_string($data) && strpos($data, "\n") !== false)
		{
			echo $pad . '<' . $parentKey . '><![CDATA[' . str_replace(']]>', ']]]]><![CDATA[>', $data) . ']]></' . $parentKey . '>' . "\n";
		}
		else
		{
			echo $pad . '<' . $parentKey . '>' . htmlspecialchars($data) . '</' . $parentKey . '>' . "\n";
		}
	}

	// ---------- General message reporting and conversion functions ----------

	// Gets the HTML attributes for CSS style and class and additional HTML attributes from a
	// combined style specification.
	//
	// Examples:
	// * 'margin-right: 5px;'
	// * '|monospace'
	// * 'margin-right: 5px;|monospace'
	// * '||disabled'
	// * 'margin-right: 5px;||disabled'
	// * 'margin-right: 5px;|monospace|disabled'
	//
	// style = (string) CSS style attribute value, CSS class attribute value and HTML attribute flags, each separated by '|'
	// addStyle = (string) Additional locally defined style and class values, same format as style parameter
	//
	// Returns: (string) HTML attributes that have a value set. If any attribute is set, there is
	//                   a single space around the returned string so it can be safely printed into
	//                   HTML elements.
	//
	private static function GetStyleClassAttr($style, $addStyle = '')
	{
		if (!isset($style)) $style = '';
		$style = trim($style);
		if (!isset($addStyle)) $addStyle = '';
		$addStyle = trim($addStyle);

		list($style, $cssClass, $attr) = explode('|', $style);
		list($addStyle, $addCssClass, $addAttr) = explode('|', $addStyle);
		$style = trim($style . ' ' . $addStyle);
		$cssClass = trim($cssClass . ' ' . $addCssClass);
		$attr = trim($attr . ' ' . $addAttr);

		$html = '';
		if (strlen($cssClass))
		{
			$html .= ' class="' . htmlspecialchars($cssClass) . '"';
		}
		if (strlen($style))
		{
			$html .= ' style="' . htmlspecialchars($style) . '"';
		}
		if (strpos($attr, 'disabled') !== false)
		{
			$html .= ' disabled="disabled"';
		}
		if (strpos($attr, 'readonly') !== false)
		{
			$html .= ' readonly="readonly"';
		}

		if (strlen($html)) $html .= ' ';
		return $html;
	}

	// Determines whether the combined style definition declares an object disabled.
	//
	// style = (string) CSS style attribute value, CSS class attribute value and HTML attribute flags, each separated by '|'
	//
	// Returns (bool)
	//
	private static function IsStyleDisabled($style)
	{
		list($style, $cssClass, $attr) = explode('|', $style);
		return strpos($attr, 'disabled') !== false;
	}

	// Begins reporting of messages.
	//
	// This must be called before the first message reported through ReportError and the other
	// functions to make use of the HTML block. That block will only be opened if at least one
	// message is generated.
	//
	public static function BeginReport()
	{
		self::$useMessageBlock = true;
	}

	// Ends reporting of messages.
	//
	// This must be called after the last message reported through ReportError and the other
	// functions to close the HTML block.
	//
	public static function EndReport()
	{
		if (self::$messageBlockOpen)
		{
			echo '</div>';   // .overlay-messages
			self::$messageBlockOpen = false;
		}
		self::$useMessageBlock = false;
	}

	// Opens the message reporting HTML block if enabled and not done yet.
	//
	private static function EnsureReportBlock()
	{
		if (self::$useMessageBlock && !self::$messageBlockOpen)
		{
			#echo '<div class="overlay-messages" onclick="DOM.hide(this)">';
			echo '<div class="overlay-messages">';
			self::$messageBlockOpen = true;
		}
	}

	// Reports an information message.
	//
	// msg = (string) Message text in plain text format (i.e. no HTML)
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ReportInfo($msg, $style = '|informationbox')
	{
		self::EnsureReportBlock();
		echo '<p' . self::GetStyleClassAttr($style) . '><strong>' . nl2br(htmlspecialchars($msg)) . '</strong></p>';
	}

	// Reports a success message.
	//
	// msg = (string) Message text in plain text format (i.e. no HTML);
	//                {{null}} to display and forget the message from the session variable SuccessMessage
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ReportSuccess($msg = null, $style = '|successbox')
	{
		if (!isset($msg))
		{
			$msg = UnbSession::GetVar('SuccessMessage', true);
			if (!isset($msg) || !strlen($msg))
				return;
		}

		self::EnsureReportBlock();
		echo '<p' . self::GetStyleClassAttr($style) . '><strong>' . nl2br(htmlspecialchars($msg)) . '</strong></p>';
	}

	// Reports an error message.
	//
	// param = (string) Message text in plain text format (i.e. no HTML)
	//         (Exception) Any type of exception to be reported
	//         {{null}} to display the error message from handling the form, if any
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ReportError($param = null, $style = '|errorbox')
	{
		if (!isset($param))
		{
			if (!self::IsFormError())
				return;
			$param = self::GetFormErrorMessage();
		}

		self::EnsureReportBlock();
		if ($param instanceof Exception)
		{
			echo '<p' . self::GetStyleClassAttr($style) . '>';
			echo UnbLocale::Translate('msg:problem processing request') . '<br />';
			echo '<strong>' . UnbLocale::Translate('exception:' . get_class($param), true) . UnbLocale::Colon() . ' ' . nl2br(htmlspecialchars($param->getMessage())) . '</strong>';
			echo '</p>';
		}
		else
		{
			echo '<p' . self::GetStyleClassAttr($style, 'white-space: pre-wrap;') . '>';
			echo UnbLocale::Translate('msg:problem processing request') . '<br />';
			foreach (explode("\n", $param) as $line)
			{
				if ($line{0} !== ' ') echo '<strong>';
				echo htmlspecialchars(rtrim($line));
				if ($line{0} !== ' ') echo '</strong>';
				echo '<br />';
			}
			echo '</p>';
		}
	}

	// Redirects the browser to another URL and terminates programme execution.
	//
	// This requires output buffering to be active to clear any previously printed output. Output
	// buffering is normally enabled in UnbEnvironment::EnableHttpCompression, called by
	// UnbEnvironment::SetupEnvironment, called by UnbUi::Init.
	//
	// url = (string) URL to redirect the browser to
	// params = (array(name => value)) Associative array of URL parameters to append to the URL
	// hash = (string) Page anchor
	//
	public static function RedirectAndExit($url, $params = null, $hash = null)
	{
		if (self::DoRedirects)
		{
			// Discard any buffered output
			// (This may also forget about GZip compression, so it needs to be re-enabled again.)
			while (ob_get_level())
				ob_end_clean();

			// Make sure compression is active if the GZip header was already sent
			foreach (headers_list() as $h)
			{
				if (preg_match('_Content-Encoding:\s*gzip_i', $h))
				{
					ob_start('ob_gzhandler');
					break;
				}
			}
		}

		if (isset($params)) foreach ($params as $name => $value)
		{
			$url .= (strpos($url, '?') !== false ? '&' : '?') .
				$name . '=' . urlencode($value);
		}
		if (isset($hash))
		{
			$url .= '#' . urlencode($hash);
		}

		if (self::DoRedirects)
		{
			header('HTTP/1.1 303 See other');
			header('Location: ' . $url);
		}

		// Basic fallback if browser redirection is disabled
		if (self::DoRedirects)
		{
			echo '<html><head><title>' . UnbLocale::Translate('msg:page redirection') . '</title>';
			echo '<meta http-equiv="refresh" content="0; URL=' . htmlspecialchars($url) . '" />';
			echo '</head><body>';
			echo '<a href="' . htmlspecialchars($url) . '">' . UnbLocale::Translate('msg:page redirection click here') . '</a>';
			echo '</body></html>';
		}
		else
		{
			echo '<p><strong><a href="' . htmlspecialchars($url) . '">Redirection has been disabled – click here to continue.</a></strong></p>';
		}

		exit();
	}

	// Prints a relative time description with an absolute time as tooltip.
	//
	// The second parameter may be left out if ret is meant to be set {{true}}.
	//
	// time = (string) Time specification
	//        (int) UNIX timestamp
	// context = (UnbTimeContext) Specifies the context in which the time will be used, used for the correct grammatical formulation
	// ret = (bool) Specifies whether the HTML code shall be returned instead of printed
	//
	public static function RelativeTime($time, $context = UnbTimeContext::Point, $ret = false)
	{
		// Accept to leave out the second parameter, recognising by the type
		if ($context === true)
		{
			$context = UnbTimeContext::Point;
			$ret = true;
		}

		$str = '<span title="' . UnbLocale::Time($time, UnbTimeFormat::DateTimeSec, true) . '">' .
			UnbLocale::RelativeTime($time, $context, true) . '</span>';
		if ($ret)
			return $str;
		else
			echo $str;
	}

	// Gets a multi-line string of user names from a list of user IDs.
	//
	// keys = (array(int)) List of user IDs
	// includeSystem = (bool) {{true}} to include the system-defined keys, {{false}} otherwise
	//
	public static function GetUserList($keys, $includeSystem = true)
	{
		if (!is_array($keys))
			throw new InvalidArgumentException('Invalid argument type: $keys');

		$list = '';
		foreach ($keys as $key)
		{
			if (!$includeSystem && $key <= 9) continue;   // Don't list system keys
			$keyUser = UnbUser::FindByUserId($key);
			if (isset($keyUser))
				$list .= $keyUser->GetDisplayName() . "\n";
			else
				$list .= '#' . $key . "\n";
		}
		return rtrim($list);
	}

	// ---------- General form support functions ----------

	// Prints the opening tag of the HTML form and the URL key hidden field.
	//
	// actionUrl = (string) Form action URL as plaintext (i.e. not HTML-encoded)
	// enableUploads = (bool) {{true}} to specify a form encoding that allows file uploads, {{false}} to use the default encoding that uses less space for normal form field values
	//
	public static function BeginForm($actionUrl = '', $enableUploads = false, $id = null)
	{
		if (!is_string($actionUrl))
			throw new InvalidArgumentException('Invalid argument type: $actionUrl');
		$actionUrl = trim($actionUrl);
		$id = self::GetId($id, 'unbForm');

		echo '<form id="' . $id . '" action="' . htmlspecialchars($actionUrl) . '" method="post"' .
			($enableUploads ? ' enctype="multipart/form-data"' : '') . '>';
		self::FormUrlKey();
	}

	// Prints the closing tag of the HTML form.
	//
	public static function EndForm()
	{
		echo '</form>';
	}

	// Prints the URL key to be submitted with the form.
	//
	public static function FormUrlKey()
	{
		echo '<input type="hidden" name="urlkey" value="' . UnbSession::GetUrlKey() . '" />';
	}

	// Prints the user ID to be submitted with the form.
	//
	public static function FormUser($userId)
	{
		if (!is_int($userId))
			throw new InvalidArgumentException('Invalid argument type: $userId');
		if ($userId <= 0)
			throw new InvalidArgumentException('Argument out of range: $userId');

		echo '<input type="hidden" name="userid" value="' . $userId . '" />';
	}

	// Prints the message revision ID to be submitted with the form.
	//
	public static function FormMsgRev($messageRevisionId)
	{
		if (!is_int($messageRevisionId))
			throw new InvalidArgumentException('Invalid argument type: $messageRevisionId');
		if ($messageRevisionId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageRevisionId');

		echo '<input type="hidden" name="messagerevisionid" value="' . $messageRevisionId . '" />';
	}

	// Prints the message ID to be submitted with the form.
	//
	public static function FormMsg($messageId)
	{
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');

		echo '<input type="hidden" name="messageid" value="' . $messageId . '" />';
	}

	// Prints the in-reply-to message ID to be submitted with the form.
	//
	public static function FormReplyToMsg($messageId)
	{
		if (isset($messageId))
		{
			if (!is_int($messageId))
				throw new InvalidArgumentException('Invalid argument type: $messageId');
			if ($messageId < 0)
				throw new InvalidArgumentException('Argument out of range: $messageId');
			if ($messageId == 0)
				$messageId = null;
		}

		if (isset($messageId))
		{
			echo '<input type="hidden" name="replyto_messageid" value="' . $messageId . '" />';
		}
	}

	// Prints the message revision ID, to which the created file shall be added to, to be
	// submitted with the form.
	//
	public static function FormFileAddToMsgRev($messageRevisionId)
	{
		if (isset($messageRevisionId))
		{
			if (!is_int($messageRevisionId))
				throw new InvalidArgumentException('Invalid argument type: $messageRevisionId');
			if ($messageRevisionId < 0)
				throw new InvalidArgumentException('Argument out of range: $messageRevisionId');
			if ($messageRevisionId == 0)
				$messageRevisionId = null;
		}

		if (isset($messageRevisionId))
		{
			echo '<input type="hidden" name="addfileto_revisionid" value="' . $messageRevisionId . '" />';
		}
	}

	// Prints the tag ID to be submitted with the form.
	//
	public static function FormTag($tagId)
	{
		if (!is_int($tagId))
			throw new InvalidArgumentException('Invalid argument type: $tagId');
		if ($tagId <= 0)
			throw new InvalidArgumentException('Argument out of range: $tagId');

		echo '<input type="hidden" name="tagid" value="' . $tagId . '" />';
	}

	// Prints JavaScript code to handle a default message for empty input fields and a dedicated hidden
	// input field that tells the form handler to ignore the default message and that the input field
	// is actually empty.
	//
	// When the default message is set, the CSS class {{default-message}} is applied to the input element.
	//
	// The JavaScript code generated here relies on dynamic-lib to be loaded.
	// See <http://unclassified.de/go/dynamic-lib>.
	//
	// id = (string) HTML element ID of the input element to generate code for
	// defaultMessage = (string) Default message to show if the input element is left empty
	// emptyFieldName = (string) Form element name of the field to set when the default message has been put in the input element. Ignored if empty. Do not use any " (double quote) here, it will break the JS code.
	//
	private static function InsertDefaultMessage($id, $defaultMessage, $emptyFieldName = '')
	{
		// The outer function block creates a local scope that contains all locally defined variables.
		// It is not stored anywhere but only executed immediately.
		echo '<script type="text/javascript">//<![CDATA[
		(function()
		{
			var o = $("#' . $id . '");
			var ef = $("input[name=\"' . $emptyFieldName . '\"]" , o.form);
			DOMEvents.add(o, "focus", function (e)
			{
				if (DOM.hasClass(e.target, "default-message"))
				{
					e.target.value = "";
					DOM.removeClass(e.target, "default-message");
					if (ef) ef.value = "0";
				}
			});
			var blurFn = function (e)
			{
				if (e.target.value.trim() == "" ||
					ef && ef.value == "1")
				{
					DOM.addClass(e.target, "default-message");
					e.target.value = "' . str_replace('"', '\\"', $defaultMessage) . '";
					if (ef) ef.value = "1";
				}
			};
			DOMEvents.add(o, "blur", blurFn);
			blurFn({target: o});
		})();
		//]]></script>';
	}

	// Does centralised HTML input element ID checking.
	//
	private static function GetId($id, $defaultId)
	{
		if (isset($id))
		{
			if (!is_string($id))
				throw new InvalidArgumentException('Invalid argument type: $id');
			$id = trim($id);
			if (!strlen($id))
				$id = null;
		}
		if (!isset($id))
		{
			$id = $defaultId;
		}
		return $id;
	}

	// Gets a field value either by a previous POST submitting or by a default value.
	//
	private static function GetValue($fieldName, $defaultValue = '', $emptyFieldName = null)
	{
		if ($_POST[$emptyFieldName])
			return '';
		return htmlspecialchars($_POST ? $_POST[$fieldName] : $defaultValue);
	}

	// ---------- User, login and registration functions ----------

	// Prints the input element for the user login name.
	//
	// user = (UnbUser) User instance to be altered, or {{null}} for a new user or login
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	// autoComplete = (bool) Enable auto completion for this field
	// enabled = (bool) Enable text input in this field
	//
	public static function UserLoginName($user = null, $id = null, $style = '', $autoComplete = true, $enabled = true)
	{
		$defaultValue = '';
		if (isset($user))
		{
			if (!($user instanceof UnbUser))
				throw new InvalidArgumentException('Invalid argument type: $user');
			$defaultValue = $user->GetLoginName();
		}
		$id = self::GetId($id, 'userLoginNameTb');

		$acAttr = '';
		if (!$autoComplete)
			$acAttr = ' autocomplete="off" ';

		$disAttr = '';
		if (!$enabled)
			$disAttr = ' disabled="disabled" ';

		echo '<input type="text" id="' . $id . '" name="user_loginname" size="20" ' . self::GetStyleClassAttr($style) . ' value="' . self::GetValue('user_loginname', $defaultValue) . '" ' . $acAttr . $disAttr . '/>
			<input type="hidden" name="action[]" value="set_user_loginname" />';
	}

	// Prints the input element for the user login password.
	//
	// user = (UnbUser) User instance to be altered, or {{null}} for a new user or login
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	// autoComplete = (bool) Enable auto completion for this field
	//
	public static function UserLoginPassword($user = null, $id = null, $style = '', $autoComplete = true)
	{
		$defaultValue = '';
		if (isset($user))
		{
			if (!($user instanceof UnbUser))
				throw new InvalidArgumentException('Invalid argument type: $user');
			$defaultValue = $user->GetLoginPassword() ? '          ' : '';
		}
		$id = self::GetId($id, 'userLoginPasswordTb');

		$acAttr = '';
		if (isset($user) || $_POST || !$autoComplete)
			$acAttr = ' autocomplete="off" ';

		echo '<input type="password" id="' . $id . '" name="user_loginpassword" size="20" ' . self::GetStyleClassAttr($style) . ' value="' . self::GetValue('user_loginpassword', $defaultValue) . '" ' . $acAttr . '/>
			<input type="hidden" name="action[]" value="set_user_loginpassword" />';
	}

	// Prints the input element for the user login password repetition.
	//
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function UserRepeatPassword($id = null, $style = '')
	{
		$id = self::GetId($id, 'userRepeatPasswordTb');

		echo '<input type="password" id="' . $id . '" autocomplete="off" name="user_repeatpassword" size="20" ' . self::GetStyleClassAttr($style) . ' value="' . self::GetValue('user_repeatpassword') . '" />';
	}

	// Prints the input element for the current user login password validation.
	//
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function UserCurrentPassword($id = null, $style = '')
	{
		$id = self::GetId($id, 'userCurrentPasswordTb');

		echo '<input type="password" id="' . $id . '" autocomplete="off" name="user_currentpassword" size="20" ' . self::GetStyleClassAttr($style) . ' value="' . self::GetValue('user_currentpassword') . '" />';
	}

	// Prints the input element for the user display name.
	//
	// user = (UnbUser) User instance to be altered, or {{null}} for a new user
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	// autoComplete = (bool) Enable auto completion for this field
	//
	public static function UserDisplayName($user = null, $id = null, $style = '', $autoComplete = true)
	{
		$defaultValue = '';
		if (isset($user))
		{
			if (!($user instanceof UnbUser))
				throw new InvalidArgumentException('Invalid argument type: $user');
			$defaultValue = $user->GetDisplayName();
		}
		$id = self::GetId($id, 'userDisplayNameTb');

		$acAttr = '';
		if (!$autoComplete)
			$acAttr = ' autocomplete="off" ';

		echo '<input type="text" id="' . $id . '" name="user_displayname" size="25" ' . self::GetStyleClassAttr($style) . ' value="' . self::GetValue('user_displayname', $defaultValue) . '" ' . $acAttr . '/>';
		echo '<input type="hidden" name="action[]" value="set_user_displayname" />';
	}

	// Gets a value indicating whether the CAPTCHA validation shall be used.
	//
	// Returns (bool)
	//
	public static function IsCaptchaEnabled()
	{
		return UnbConfiguration::Get('captcha.enabled', true);
	}

	// Prints the image and input element for the CAPTCHA test.
	//
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// text1 = (string) Main descriptive text (will not be HTML-encoded, put <br /> at the end), or {{null}} for built-in default text
	// text2 = (string) Second command text (will not be HTML-encoded, put <br /> at the end), or {{null}} for built-in default text
	//
	public static function Captcha($id = null, $text1 = null, $text2 = null)
	{
		$id = self::GetId($id, 'captchaTb');
		if (!isset($text1))
			$text1 = UnbLocale::Translate('msg:captcha 1') . '<br />';
		if (!isset($text2))
			$text2 = UnbLocale::Translate('msg:captcha 2') . '<br />';

		echo $text1 . '
			<img src="captcha.php?r=' . mt_rand(1, 9999999) . '" alt="CAPTCHA image" style="margin: 0.2em 0;" /><br />
			' . $text2 . '
			<img src="' . self::$data['img_url'] . 'btn-downhere.png" alt="" />
			<input type="text" id="' . $id . '" name="captcha_word" autocomplete="off" size="15" value="' . self::GetValue('captcha_word') . '" />';
	}

	// Gets a user's additional keys as multi-line string.
	//
	// user = (UnbUser) User instance to get the key list from
	//
	// Returns (string) All additional keys with their display names, one per line, separated by \n
	//
	public static function GetUserAdditionalKeys($user)
	{
		if (!($user instanceof UnbUser))
			throw new InvalidArgumentException('Invalid argument type: $user');

		return self::GetUserList($user->GetAdditionalKeys(), false);
	}

	// Prints the input element for the user additional keys list.
	//
	// user = (UnbUser) User instance to be altered, or {{null}} for a new user
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function UserAdditionalKeysList($user = null, $id = null, $style = '')
	{
		$additionalKeysList = '';
		if (isset($user))
		{
			if (!($user instanceof UnbUser))
				throw new InvalidArgumentException('Invalid argument type: $user');
			$additionalKeysList = self::GetUserAdditionalKeys($user);
		}
		$id = self::GetId($id, 'userAdditionalKeysListTb');

		echo '<textarea id="' . $id . '" name="user_additionalkeyslist" cols="30" rows="3" ' . self::GetStyleClassAttr($style) . '>' . self::GetValue('user_additionalkeyslist', $additionalKeysList, 'user_additionalkeys_empty') . '</textarea>
			<input type="hidden" name="user_additionalkeys_empty" value="0" />
			<input type="hidden" name="action[]" value="set_user_additionalkeys" />';
	}

	// Prints the input element for the user's Administrator privilege.
	//
	// user = (UnbUser) User instance to be altered, or {{null}} for a new user
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	//
	public static function UserAdministrator($user = null, $id = null)
	{
		$defaultValue = false;
		if (isset($user))
		{
			if (!($user instanceof UnbUser))
				throw new InvalidArgumentException('Invalid argument type: $user');
			$defaultValue = $user->IsAdministrator();
		}
		$id = self::GetId($id, 'userAdministratorCb');

		echo '<label><input type="checkbox" id="' . $id . '" name="user_is_administrator" value="1" ' . (($_POST ? $_POST['user_is_administrator'] : $defaultValue) ? 'checked="checked"' : '') . ' /> ' . UnbLocale::Translate('user:administrator') . '</label>
			<input type="hidden" name="action[]" value="set_user_administrator" />';
	}

	// Prints the input element for the user's Moderator privilege.
	//
	// user = (UnbUser) User instance to be altered, or {{null}} for a new user
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	//
	public static function UserModerator($user = null, $id = null)
	{
		$defaultValue = false;
		if (isset($user))
		{
			if (!($user instanceof UnbUser))
				throw new InvalidArgumentException('Invalid argument type: $user');
			$defaultValue = $user->IsModerator();
		}
		$id = self::GetId($id, 'userModeratorCb');

		echo '<label><input type="checkbox" id="' . $id . '" name="user_is_moderator" value="1" ' . (($_POST ? $_POST['user_is_moderator'] : $defaultValue) ? 'checked="checked"' : '') . ' /> ' . UnbLocale::Translate('user:moderator') . '</label>
			<input type="hidden" name="action[]" value="set_user_moderator" />';
	}

	// Prints the input element for the user's Trusted privilege.
	//
	// user = (UnbUser) User instance to be altered, or {{null}} for a new user
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	//
	public static function UserTrusted($user = null, $id = null)
	{
		$defaultValue = false;
		if (isset($user))
		{
			if (!($user instanceof UnbUser))
				throw new InvalidArgumentException('Invalid argument type: $user');
			$defaultValue = $user->IsTrusted();
		}
		$id = self::GetId($id, 'userTrustedCb');

		echo '<label><input type="checkbox" id="' . $id . '" name="user_is_trusted" value="1" ' . (($_POST ? $_POST['user_is_trusted'] : $defaultValue) ? 'checked="checked"' : '') . ' /> ' . UnbLocale::Translate('user:trusted') . '</label>
			<input type="hidden" name="action[]" value="set_user_trusted" />';
	}

	// Prints the input element for a user vcard entry.
	//
	// type = (string) VCard entry type
	// entry = (array('Data' => ..., 'Visibility' => ...)) VCard entry as returned by UnbUser::GetVcardEntries
	//
	public static function UserAlterVcardEntry($type, $entry)
	{
		if (!self::$data['action_set_user_vcard'])
		{
			echo '<input type="hidden" name="action[]" value="set_user_vcard" />';
			self::$data['action_set_user_vcard'] = true;
		}

		echo '<input type="text" name="vcard_data_' . $type . '" size="30" value="' . self::GetValue('vcard_data_' . $type, $entry['Data']) . '" /> &nbsp;
			' . UnbLocale::Translate('user-vcard:visibility') . UnbLocale::Colon() . ' <select name="vcard_visibility_' . $type . '">
				<option value="' . UnbUserVcardVisibility::Public_ . '" ' . (($_POST ? $_POST['vcard_visibility_' . $type] : $entry['Visibility']) == UnbUserVcardVisibility::Public_ ? 'selected="selected"' : '') . '>' . UnbLocale::Translate('user-vcard:public') . '</option>
				<option value="' . UnbUserVcardVisibility::Internal . '" ' . (($_POST ? $_POST['vcard_visibility_' . $type] : $entry['Visibility']) == UnbUserVcardVisibility::Internal ? 'selected="selected"' : '') . '>' . UnbLocale::Translate('user-vcard:internal') . '</option>
				<option value="' . UnbUserVcardVisibility::Confidential . '" ' . (($_POST ? $_POST['vcard_visibility_' . $type] : $entry['Visibility']) == UnbUserVcardVisibility::Confidential ? 'selected="selected"' : '') . '>' . UnbLocale::Translate('user-vcard:confidential') . '</option>
				<option value="' . UnbUserVcardVisibility::Private_ . '" ' . (($_POST ? $_POST['vcard_visibility_' . $type] : $entry['Visibility']) == UnbUserVcardVisibility::Private_ ? 'selected="selected"' : '') . '>' . UnbLocale::Translate('user-vcard:private') . '</option>
			</select> &nbsp;
			<label><input type="checkbox" name="vcard_remove_' . $type . '" value="1" ' . ($_POST['vcard_remove_' . $type] ? 'checked="checked"' : '') . ' /> ' . UnbLocale::Translate('user-vcard:remove this entry') . '</label>';
	}

	// Prints the input element for a new user vcard entry type.
	//
	// types = (array) Types to select from
	//
	public static function UserVcardTypes($types)
	{
		echo '<select name="vcard_newtype">';
		array_unshift($types, '');
		foreach ($types as $type)
		{
			$label = UnbLocale::Translate('vcard:' . mb_strtolower($type), true);
			echo '<option value="' . $type . '" ' . ($type == $_POST['vcard_newtype'] ? ' selected="selected"' : '') . '>' . $type . '</option>';
		}
		echo '</select>';
	}

	// Prints the input element for a new user vcard entry.
	//
	public static function UserAddVcardEntry()
	{
		if (!self::$data['action_set_user_vcard'])
		{
			echo '<input type="hidden" name="action[]" value="set_user_vcard" />';
			self::$data['action_set_user_vcard'] = true;
		}

		echo '<input type="text" name="vcard_newdata" size="30" value="' . self::GetValue('vcard_newdata') . '" /> &nbsp;
			' . UnbLocale::Translate('user-vcard:visibility') . UnbLocale::Colon() . ' <select name="vcard_newvisibility">
				<option value="' . UnbUserVcardVisibility::Public_ . '" ' . ($_POST['vcard_newvisibility'] == UnbUserVcardVisibility::Public_ ? 'selected="selected"' : '') . '>' . UnbLocale::Translate('user-vcard:public') . '</option>
				<option value="' . UnbUserVcardVisibility::Internal . '" ' . ($_POST['vcard_newvisibility'] == UnbUserVcardVisibility::Internal ? 'selected="selected"' : '') . '>' . UnbLocale::Translate('user-vcard:internal') . '</option>
				<option value="' . UnbUserVcardVisibility::Confidential . '" ' . ($_POST['vcard_newvisibility'] == UnbUserVcardVisibility::Confidential ? 'selected="selected"' : '') . '>' . UnbLocale::Translate('user-vcard:confidential') . '</option>
				<option value="' . UnbUserVcardVisibility::Private_ . '" ' . ($_POST['vcard_newvisibility'] == UnbUserVcardVisibility::Private_ ? 'selected="selected"' : '') . '>' . UnbLocale::Translate('user-vcard:private') . '</option>
			</select>';
	}

	// Prints the input element for the option to reset the user's autologin password.
	//
	// user = (UnbUser) User instance to be altered
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	//
	public static function UserResetAutologinPassword($user, $id = null)
	{
		if (!($user instanceof UnbUser))
			throw new InvalidArgumentException('Invalid argument type: $user');
		$id = self::GetId($id, 'userResetAutologinPwdCb');

		echo '<label><input type="checkbox" id="' . $id . '" name="reset_user_autologin_password_value" value="1" ' . (($_POST ? $_POST['reset_user_autologin_password_value'] : $defaultValue) ? 'checked="checked"' : '') . ' /> ' . UnbLocale::Translate('user:reset auto-login password') . '</label>
			<input type="hidden" name="action[]" value="reset_user_autologin_password" />';
	}

	// Prints the HTML link to approve a message revision.
	//
	// user = (UnbUser) User instance to be altered
	// htmlTemplate = (string) String to be printed, with "%s" replaced by URL query parameters
	//
	public static function UserResetAutologinPasswordLink($user, $htmlTemplate)
	{
		if (!($user instanceof UnbUser))
			throw new InvalidArgumentException('Invalid argument type: $user');
		if (!is_string($htmlTemplate))
			throw new InvalidArgumentException('Invalid argument type: $htmlTemplate');

		echo str_replace(
			'%s',
			'action=reset_user_autologin_password&amp;urlkey=' . UnbSession::GetUrlKey() . '&amp;userid=' . $user->GetUserId(),
			$htmlTemplate);
	}

	// Prints the login controls.
	//
	// loginLink = (string) URL to the login page, displayed if not logged in
	// userLink = (string) URL to the user account page. "%i" is replaced by the logged in user ID.
	// logoutLink = (string) URL to the logout page. "%k" is replaced by the URL key.
	// logoutWarningLink = (string) URL to the logout page warning the user about losing account access
	//
	public static function UserLoginLink($loginLink, $userLink, $logoutLink, $logoutWarningLink)
	{
		$sessionUser = UnbSession::GetSessionUser();
		if (isset($sessionUser))
		{
			$userLink = str_replace(
				'%i',
				$sessionUser->GetUserId(),
				$userLink);
			$logoutLink = str_replace(
				'%k',
				UnbSession::GetUrlKey(),
				$logoutLink);

			echo UnbLocale::Capital(UnbLocale::Translate('session:logged in as')) . UnbLocale::Colon() .
				' <strong><a href="' . htmlspecialchars($userLink) . '">' .
				htmlspecialchars($sessionUser->GetDisplayName()) .
				'</a></strong>';
			#$statusStr = '';
			#if ($sessionUser->IsAdministrator())
			#	$statusStr .= ($statusStr ? ', ' : '') . 'Admin';
			#if ($sessionUser->IsModerator())
			#	$statusStr .= ($statusStr ? ', ' : '') . 'Mod';
			#if ($sessionUser->IsTrusted())
			#	$statusStr .= ($statusStr ? ', ' : '') . 'Trusted';
			#if ($statusStr)
			#	echo ' <small>(' . $statusStr . ')</small>';

			if (!strlen($sessionUser->GetLoginName()) ||
				!strlen($sessionUser->GetLoginPassword()))
			{
				echo ' &nbsp;–&nbsp; <a href="' . htmlspecialchars($logoutWarningLink) . '">' .
					UnbLocale::Capital(UnbLocale::Translate('session:log out warning')) . '</a>';
			}
			else
			{
				echo ' &nbsp;–&nbsp; <a href="' . htmlspecialchars($logoutLink) . '">' .
					UnbLocale::Capital(UnbLocale::Translate('session:log out')) . '</a>';
			}
		}
		else
		{
			echo UnbLocale::Capital(UnbLocale::Translate('session:not logged in'));
			echo ' &nbsp;–&nbsp; <a href="' . htmlspecialchars($loginLink) . '">';
			echo UnbLocale::Capital(UnbLocale::Translate('session:log in or register'));
			echo '</a>';
		}
	}

	// ---------- Message and message list display functions ----------

	// Prints a message reference tree.
	//
	// messageId = (int) Message ID to show the tree for. This doesn't need to be the root message if $autoRoot is set.
	// rootMessageId = (int) Root message of the discussion. Automatically select the discussion root message to start the tree with if unset.
	// autoSelect = (bool) Automatically select other messages in the tree for display
	// maxLevel = (int) Number of reply levels to load and display
	// htmlTemplate = (string) String to be printed as link element (<a>).
	//                         * "%q" is replaced by the URL query parameters
	//                         * "%i" is replaced by the message ID
	//                         * "%s" is replaced by the subject
	// moreLink = (string) URL to display more replies in a deeper level.
	//                         * "%i" is replaced by the selected message ID
	//                         * "%r" is replaced by the root message ID
	//
	public static function MsgTree($messageId, $rootMessageId, $autoSelect, $maxLevel, $htmlTemplate, $moreLink)
	{
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');

		$items = UnbSearch::GetMessageTree($messageId, $rootMessageId, $autoSelect, $maxLevel);
		self::$treeMessages =& $items;

		$prevLevel = 0;
		$subjects = array();
		$levelWarning = false;
		foreach ($items as $item)
		{
			$level = $item['level'];
			$msg =& $item['msg'];
			$rev =& $item['rev'];
			$selected = $item['selected'];
			$morecount = $item['morecount'];

			$msgFlags = $msg->GetModerationFlags();
			$msgCreatedTime = $msg->GetCreatedTime();

			while ($level > $prevLevel)
			{
				echo '<div style="border-left: dotted 1px #dddddd; padding-left: 1.2em;">';
				$prevLevel++;
			}
			while ($level < $prevLevel)
			{
				echo '</div>';
				$prevLevel--;
			}

			echo '<div class="messageentry' . ($selected ? ' selected' : '') . '">';

			if ($level == 0 && !$msg->IsEntryPoint() && $msg->GetPrimaryReference())
			{
				$count = 0;
				$refMsg = $msg;
				for ($i = $maxLevel; $i > 0 && !$refMsg->IsEntryPoint() && $refMsg->GetPrimaryReference(); $i--)
				{
					$refId = $refMsg->GetPrimaryReference();
					$refMsg = UnbMessage::FindByMessageId($refId);
					if (!isset($refMsg))
						throw new Exception('The message ID ' . $refId . ' does not exist (database integrity error).');
					$count++;
				}
				$extRefMsg = $refMsg;
				while (!$extRefMsg->IsEntryPoint() && $extRefMsg->GetPrimaryReference())
				{
					$refId = $extRefMsg->GetPrimaryReference();
					$extRefMsg = UnbMessage::FindByMessageId($refId);
					if (!isset($extRefMsg))
						throw new Exception('The message ID ' . $refId . ' does not exist (database integrity error).');
					$count++;
				}

				$url = $moreLink;
				$url = str_replace(
					'%i',
					$messageId,
					$url);
				$url = str_replace(
					'%r',
					$refMsg->GetMessageId(),
					$url);
				echo '<div class="previous">';
				echo '<a href="' . htmlspecialchars($url) . '">« ' . UnbLocale::Capital(UnbLocale::Translate('msg:references', $count)) . '</a>';
				echo '</div>';
			}

			echo '<div class="author-line">';
			$html = $htmlTemplate;
			$html = str_replace(
				'%q',
				'messageid=' . ($selected ? $messageId : $msg->GetMessageId()),
				$html);
			$html = str_replace(
				'%i',
				$msg->GetMessageId(),
				$html);
			$html = str_replace(
				'%s',
				$msg->GetOwnerUser()->GetDisplayNameOrId(),
				$html);
			echo $html;
			if ($rev->GetContentLevel() >= UnbMessageRevisionContentLevel::Subject &&
				strlen($rev->GetSubject()))
			{
				if ($level > 0 &&
					$rev->GetSubject() != $subjects[$level - 1] &&
					$rev->GetSubject() != 'Re: ' . $subjects[$level - 1])
				{
					echo ' &nbsp; ' . htmlspecialchars(UnbLocale::Quote($rev->GetSubject()));
				}
				$subjects[$level] = $rev->GetSubject();
			}
			else
			{
				$subjects[$level] = $subjects[$level - 1];   // Copy subject from parent level, assume it's the same
			}
			echo '</div>';   // .author-line

			echo '<div class="info-line">';
			self::RelativeTime($msgCreatedTime);
			$flags_array = array();
			if ($rev->GetModerationState() == UnbMessageRevisionModerationState::WaitingForApproval)
				$flags_array[] = UnbLocale::Translate('msg:unapproved');
			if ($msgFlags & UnbMessageModerationFlag::Locked || $rev->GetModerationState() == UnbMessageRevisionModerationState::Locked)
				$flags_array[] = UnbLocale::Translate('msg:locked');
			if ($msgFlags & UnbMessageModerationFlag::Hidden)
				$flags_array[] = UnbLocale::Translate('msg:hidden');
			if ($flags_array)
				echo '<br /><span class="modwarning">' . UnbLocale::Capital(join(', ', $flags_array)) . '</span>';
			echo '</div>';   // .info-line

			if ($morecount)
			{
				$url = $moreLink;
				$url = str_replace(
					'%i',
					$msg->GetMessageId(),
					$url);
				$url = str_replace(
					'%r',
					$msg->GetMessageId(),
					$url);
				echo '<div class="next">';
				echo '<a href="' . htmlspecialchars($url) . '">' . UnbLocale::Capital(UnbLocale::Translate('msg:more replies', $morecount)) . ' »</a>';
				echo '</div>';
			}

			echo '</div>';   // .messageentry

			#if (in_array($messageId, $seenMessageIds))
			#{
			#	echo '<div style="border-left: solid 1px #DDDDDD; padding-left: 1.5em;">';
			#	echo '<em style="color: red;">Reference loop!</em>';
			#	echo '</div>';
			#	return;
			#}
		}
		while ($prevLevel--)
		{
			echo '</div>';
		}

		if (count($items) <= 1)
		{
			echo '<div><em>' . UnbLocale::Capital(UnbLocale::Translate('msg:no reply')) . '</em></div>';
		}

		// Post-process to provide more data about the replies structure
		foreach ($items as &$item)
		{
			// Only process messages selected in the tree
			if ($item['selected'])
			{
				// Check for multiple replies to the previous message, on the same level as this
				$replies = array();
				foreach ($items as &$item2)
				{
					if ($item2['msg']->GetPrimaryReference() == $item['msg']->GetPrimaryReference())
					{
						$replies[] = &$item2;
					}
				}
				$item['replies'] = $replies;
				if (count($replies) > 1)
				{
					// List each reply message
					foreach ($replies as &$reply)
					{
						$msg = $reply['msg'];
						$rev = $reply['rev'];

						$counting = false;
						$minLevel = 0;
						$count = 0;
						#echo 'Replies besides #' . $msg->GetMessageId() . ': ';   // DEBUG
						foreach ($items as &$item2)
						{
							if ($item2['msg']->GetMessageId() == $msg->GetMessageId())
							{
								// Found the current reply in the tree, start counting
								$counting = true;
								$minLevel = $item2['level'];
								$count++;
								#echo '#' . $item2['msg']->GetMessageId() . ' ';
								if ($item2['morecount'])
								{
									$count += $item2['morecount'];
									#echo '+' . $item2['morecount'] . ' ';
								}
							}
							else if ($counting && $item2['level'] <= $minLevel)
							{
								// Reached the starting level, stop counting for this reply
								$counting = false;
							}
							else if ($counting)
							{
								// Counting...
								$count++;
								#echo '#' . $item2['msg']->GetMessageId() . ' ';
								if ($item2['morecount'])
								{
									$count += $item2['morecount'];
									#echo '+' . $item2['morecount'] . ' ';
								}
							}
						}
						$reply['replies-count'] = $count;
					}
				}
			}
		}

	}

	// Gets all messages displayed in a previous message tree.
	//
	// Returns: An array with one item for each message. Each item is an array itself. The
	// UnbMessage object is stored in the key 'msg', the UnbMessageRevision object is stored in
	// the key 'rev'. The key 'selected' specifies whether the message is selected for full
	// display. If no tree was generated or no messages were requested to be selected, the return
	// value is {{null}}.
	//
	public static function GetTreeMessages()
	{
		return self::$treeMessages;
	}

	// Gets a message revision by its message ID and revision number for altering or replying to it.
	//
	// messageId = (int) Message ID, {{null}} to determine from messageRevisionId
	// messageRevisionId = (int) Message revision ID, {{null}} for the default search revision
	// out msg = (UnbMessage) Message object
	// out rev = (UnbMessageRevision) Message revision object
	// out err = (string) Error message in case of failure
	//
	// Returns (bool) {{true}} on success, {{false}} otherwise.
	//
	public static function GetMsgRev($messageId, $messageRevisionId = null, &$msg, &$rev, &$err, $withHtml = false)
	{
		if (isset($messageId))
		{
			if (!is_int($messageId))
				throw new InvalidArgumentException('Invalid argument type: $messageId');
			if ($messageId <= 0)
				throw new InvalidArgumentException('Argument out of range: $messageId');
		}
		if (isset($messageRevisionId))
		{
			if (!is_int($messageRevisionId))
				throw new InvalidArgumentException('Invalid argument type: $messageRevisionId');
			if ($messageRevisionId <= 0)
				throw new InvalidArgumentException('Argument out of range: $messageRevisionId');
		}
		if (!isset($messageId) && !isset($messageRevisionId))
			throw new InvalidArgumentException('Neither $messageId nor $messageRevisionId is set.');

		// Initialise the output parameters
		$msg = null;
		$rev = null;
		$err = null;

		$cl = $withHtml ? UnbMessageRevisionContentLevel::HtmlContent : UnbMessageRevisionContentLevel::Content;

		if (isset($messageRevisionId))
		{
			$rev = UnbMessageRevision::FindByRevisionId($messageRevisionId, $cl, true);
			if (!isset($rev))
			{
				$err = 'Invalid message revision ID';
				return false;
			}

			if (!isset($messageId))
			{
				$messageId = $rev->GetMessageId();
			}
			else if ($messageId != $rev->GetMessageId())
			{
				$err = 'Invalid message revision ID';
				return false;
			}
		}

		$msg = UnbMessage::FindByMessageId($messageId);
		if (!isset($msg))
		{
			$err = 'Invalid message ID';
			return false;
		}

		if (!isset($messageRevisionId))
		{
			if ($msg->IsDraft())
			{
				// This has special means of getting a draft revision if the whole message is a draft
				$rev = $msg->GetLatestRevision($cl, true);
			}
			else
			{
				$rev = $msg->GetSearchOrLatestRevision($cl, true);
			}
			if (!isset($rev))
			{
				$err = 'Invalid message revision ID';
				return false;
			}
		}
		return true;
	}

	// Gets all draft message revisions for the current user.
	//
	// Returns (array(UnbMessageRevision))
	//
	public static function GetDraftRevisions()
	{
		$draftRevIDs = UnbMessageRevision::FindAllDrafts();
		if (count($draftRevIDs))
		{
			$draftRevs = array();
			foreach ($draftRevIDs as $draftRevID)
			{
				$rev = UnbMessageRevision::FindByRevisionId($draftRevID, UnbMessageRevisionContentLevel::Subject);
				$draftRevs[] = $rev;
			}
			return $draftRevs;
		}
		return false;
	}

	// Prints all tags assigned to the specified message revision.
	//
	// rev = (UnbMessageRevision) Message revision to print the tags for
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ShowMsgTags($rev, $style = '')
	{
		$tags = $rev->GetTags(true);
		if ($tags)
		{
			echo '<span' . self::GetStyleClassAttr($style) . '>';
			$count = 0;
			foreach ($tags as $tag)
			{
				#if ($count++) echo ', ';
				echo '<span class="tag" title="' . htmlspecialchars($tag->GetDescription()) . '">';
				echo $tag->GetTagName();
				if (($tag->GetFlags() & UnbTagFlag::UserData) && ($tag->GetTagData() != null))
					echo ': ' . htmlspecialchars($tag->GetTagData());
				echo '</span> ';
			}
			echo '</span>';
		}
	}

	// Gets a quoted message subject.
	//
	// Either both $msg and $rev must be set or none at all.
	// If no message and message revision are provided, an empty subject line is returned gracefully.
	//
	// msg = (UnbMessage) Message instance to be replied to, or {{null}}
	// rev = (UnbMessageRevision) Message revision instance to be replied to, or {{null}}
	//
	// Returns (string) The quoted subject line
	//
	public static function GetQuotedSubject($msg, $rev)
	{
		if (isset($msg) != isset($rev))
			throw new InvalidArgumentException('Invalid arguments: Either both $msg and $rev must be set or none at all.');
		if (isset($msg))
		{
			if (!($msg instanceof UnbMessage))
				throw new InvalidArgumentException('Invalid argument type: $msg');
		}
		if (isset($rev))
		{
			if (!($rev instanceof UnbMessageRevision))
				throw new InvalidArgumentException('Invalid argument type: $rev');
		}

		$quotation = '';
		if (isset($msg) && isset($rev))
		{
			$content = $rev->GetSubject();
			$content = preg_replace('_^(re:\s*)+_i', '', $content);
			$quotation = 'Re: ' . $content;
		}
		return $quotation;
	}

	// Gets a quoted message content.
	//
	// Either both $msg and $rev must be set or none at all.
	// If no message and message revision are provided, an empty content is returned gracefully.
	//
	// msg = (UnbMessage) Message instance to be replied to
	// rev = (UnbMessageRevision) Message revision instance to be replied to
	//
	// Returns (string) The quoted message content
	//
	public static function GetQuotedMsg($msg, $rev)
	{
		if (isset($msg) != isset($rev))
			throw new InvalidArgumentException('Invalid arguments: Either both $msg and $rev must be set or none at all.');
		if (isset($msg))
		{
			if (!($msg instanceof UnbMessage))
				throw new InvalidArgumentException('Invalid argument type: $msg');
		}
		if (isset($rev))
		{
			if (!($rev instanceof UnbMessageRevision))
				throw new InvalidArgumentException('Invalid argument type: $rev');
		}

		$quotation = '';
		if (isset($msg) && isset($rev))
		{
			$content = $rev->GetContent();
			if (strlen($content))
			{
				$quotation =
					UnbLocale::Translate(
						'msg:quote header',
						'user', $msg->GetOwnerUser()->GetDisplayName(),
						'time', UnbLocale::RelativeTime($msg->GetCreatedTime(), UnbTimeContext::Before, false)) .
					UnbLocale::Colon() . "\n" .
					UnbMarkup::GetQuotedText($content);
			}
		}
		return $quotation;
	}

	// Prints the HTML link to approve a message revision.
	//
	// rev = (UnbMessageRevision) Message revision instance to approve
	// htmlTemplate = (string) String to be printed, with "%s" replaced by URL query parameters
	//
	public static function MsgApproveRevLink($rev, $htmlTemplate)
	{
		if (!($rev instanceof UnbMessageRevision))
			throw new InvalidArgumentException('Invalid argument type: $rev');
		if (!is_string($htmlTemplate))
			throw new InvalidArgumentException('Invalid argument type: $htmlTemplate');

		if ($rev->GetModerationState() == UnbMessageRevisionModerationState::WaitingForApproval)
		{
			echo str_replace(
				'%s',
				'action=moderate_messagerevision&amp;urlkey=' . UnbSession::GetUrlKey() . '&amp;messagerevisionid=' . $rev->GetMessageRevisionId() . '&amp;message_modstate=' . UnbMessageRevisionModerationState::Approved,
				$htmlTemplate);
		}
	}

	// Prints the HTML link to lock a message revision.
	//
	// rev = (UnbMessageRevision) Message revision instance to lock
	// htmlTemplate = (string) String to be printed, with "%s" replaced by URL query parameters
	//
	public static function MsgLockRevLink($rev, $htmlTemplate)
	{
		if (!($rev instanceof UnbMessageRevision))
			throw new InvalidArgumentException('Invalid argument type: $rev');
		if (!is_string($htmlTemplate))
			throw new InvalidArgumentException('Invalid argument type: $htmlTemplate');

		if ($rev->GetModerationState() == UnbMessageRevisionModerationState::WaitingForApproval ||
			$rev->GetModerationState() == UnbMessageRevisionModerationState::Approved)
		{
			echo str_replace(
				'%s',
				'action=moderate_messagerevision&amp;urlkey=' . UnbSession::GetUrlKey() . '&amp;messagerevisionid=' . $rev->GetMessageRevisionId() . '&amp;message_modstate=' . UnbMessageRevisionModerationState::Locked,
				$htmlTemplate);
		}
	}

	// Prints the HTML link to unlock a message revision.
	//
	// rev = (UnbMessageRevision) Message revision instance to unlock
	// htmlTemplate = (string) String to be printed, with "%s" replaced by URL query parameters
	//
	public static function MsgUnlockRevLink($rev, $htmlTemplate)
	{
		if (!($rev instanceof UnbMessageRevision))
			throw new InvalidArgumentException('Invalid argument type: $rev');
		if (!is_string($htmlTemplate))
			throw new InvalidArgumentException('Invalid argument type: $htmlTemplate');

		if ($rev->GetModerationState() == UnbMessageRevisionModerationState::Locked)
		{
			echo str_replace(
				'%s',
				'action=moderate_messagerevision&amp;urlkey=' . UnbSession::GetUrlKey() . '&amp;messagerevisionid=' . $rev->GetMessageRevisionId() . '&amp;message_modstate=' . UnbMessageRevisionModerationState::Approved,
				$htmlTemplate);
		}
	}

	// Prints the HTML link to rate a message.
	//
	// msg = (UnbMessage) Message instance to rate
	// htmlTemplate = (string) String to be printed, with "%s" replaced by URL query parameters
	// rating = (int) New message rating
	//
	public static function MsgRateLink($msg, $htmlTemplate, $rating)
	{
		if (!($msg instanceof UnbMessage))
			throw new InvalidArgumentException('Invalid argument type: $msg');
		if (!is_string($htmlTemplate))
			throw new InvalidArgumentException('Invalid argument type: $htmlTemplate');

		echo str_replace(
			'%s',
			'action=rate_message&amp;urlkey=' . UnbSession::GetUrlKey() . '&amp;messageid=' . $msg->GetMessageId() . '&amp;message_rating=' . $rating,
			$htmlTemplate);
	}

	// Prints the HTML link to clear a message rating.
	//
	// msg = (UnbMessage) Message instance to rate
	// htmlTemplate = (string) String to be printed, with "%s" replaced by URL query parameters
	//
	public static function MsgClearRatingLink($msg, $htmlTemplate)
	{
		if (!($msg instanceof UnbMessage))
			throw new InvalidArgumentException('Invalid argument type: $msg');
		if (!is_string($htmlTemplate))
			throw new InvalidArgumentException('Invalid argument type: $htmlTemplate');

		echo str_replace(
			'%s',
			'action=clear_rate_message&amp;urlkey=' . UnbSession::GetUrlKey() . '&amp;messageid=' . $msg->GetMessageId(),
			$htmlTemplate);
	}

	// Prints the HTML code to visualise a message rating.
	//
	// msg = (UnbMessage) Message instance to rate
	// minCount = (int) Minimum number of ratings for this message to display the rating
	// minWeight = (int) Minimum rating weight sum for this message to display the rating
	// width = (int) Width of the rating bar in pixels
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function MsgRating($msg, $minCount = 1, $minWeight = 30, $width = 50, $style = '')
	{
		if (!($msg instanceof UnbMessage))
			throw new InvalidArgumentException('Invalid argument type: $msg');

		$rating = $msg->GetRating();
		if ($rating[0] >= $minCount && $rating[1] >= $minWeight)
		{
			$count = $rating[0];
			$value = $rating[2];
			$width--;
			$text = UnbLocale::Translate(
				'msg:average rating',
				'rating', UnbLocale::Number($value, 2, true),
				'count', $count);
			// There must be a zero-width joiner (&zwj;) so that the span element isn't empty or it may be omitted.
			if ($value < 0)
			{
				echo '<span' . self::GetStyleClassAttr($style) . ' title="' . $text . '"><span class="negative" style="margin-left: ' . ((10 + $value) / 10 * $width / 2) . 'px; padding-left: ' . (-$value / 10 * $width / 2) . 'px; margin-right: ' . ($width / 2) . 'px;">&zwj;</span></span>';
			}
			else if ($value >= 0)
			{
				echo '<span' . self::GetStyleClassAttr($style) . ' title="' . $text . '"><span class="positive" style="margin-left: ' . ($width / 2) . 'px; padding-left: ' . ($value / 10 * $width / 2) . 'px; margin-right: ' . ((10 - $value) / 10 * $width / 2) . 'px;">&zwj;</span></span>';
			}
		}
	}

	// Gets a value indicating whether a particular message (revision) can be altered.
	//
	// For a message revision to be alterable, its content must be loaded and access control must allow
	// altering the message.
	//
	// msg = (UnbMessage) Message instance to be checked
	// rev = (UnbMessageRevision) Message revision instance to be checked
	//
	// Returns (bool) {{true}} if the message (revision) can be altered, {{false}} otherwise
	//
	public static function CanAlterMsg($msg, $rev)
	{
		if (!($msg instanceof UnbMessage))
			throw new InvalidArgumentException('Invalid argument type: $msg');
		if (!($rev instanceof UnbMessageRevision))
			throw new InvalidArgumentException('Invalid argument type: $rev');

		return $rev->GetContentLevel() >= UnbMessageRevisionContentLevel::Content && $msg->IsAlterAllowed();
	}

	// Gets a value indicating whether a particular message (revision) can be replied to.
	//
	// For a message revision to be replyable, its content must be loaded and access control must allow
	// replying to the message.
	//
	// msg = (UnbMessage) Message instance to be checked
	// rev = (UnbMessageRevision) Message revision instance to be checked
	//
	// Returns (bool) {{true}} if the message (revision) can be replied to, {{false}} otherwise
	//
	public static function CanReplyMsg($msg, $rev)
	{
		if (!($msg instanceof UnbMessage))
			throw new InvalidArgumentException('Invalid argument type: $msg');
		if (!($rev instanceof UnbMessageRevision))
			throw new InvalidArgumentException('Invalid argument type: $rev');

		return $rev->GetContentLevel() >= UnbMessageRevisionContentLevel::Content && $msg->IsReplyAllowed();
	}

	// Gets a value indicating whether the subject of a particular message revision is accessible.
	//
	// rev = (UnbMessageRevision) Message revision instance to be checked
	//
	// Returns (bool) {{true}} if the subject is accessible, {{false}} otherwise
	//
	public static function CanSeeSubject($rev)
	{
		if (!($rev instanceof UnbMessageRevision))
			throw new InvalidArgumentException('Invalid argument type: $rev');

		return $rev->GetContentLevel() >= UnbMessageRevisionContentLevel::Subject;
	}

	// Gets a value indicating whether the current session user is the owner of a particular message.
	//
	// msg = (UnbMessage) Message instance to be checked
	//
	// Returns (bool) {{true}} if the session user if the message owner, {{false}} otherwise
	//
	public static function IsMsgOwner($msg)
	{
		if (!($msg instanceof UnbMessage))
			throw new InvalidArgumentException('Invalid argument type: $msg');

		return UnbSession::GetSessionUserId() == $msg->GetOwnerUserId();
	}

	// ---------- Message editor display functions ----------

	// Prints the input element for the message subject.
	//
	// newMessage = (bool) {{true}} to create a new message (replying to another one) or {{false}} to alter a message
	// msg = (UnbMessage) Message instance to be replied to resp. to be altered, or {{null}} for an empty new message
	// rev = (UnbMessageRevision) Message revision instance to be replied to resp. to be altered, or {{null}} for an empty new message
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function MsgSubject($newMessage = true, $msg = null, $rev = null, $id = null, $style = '')
	{
		$id = self::GetId($id, 'msgSubjectTb');

		if ($newMessage)
		{
			if (!self::$data['action_create_message_set'])
			{
				echo '<input type="hidden" name="action[]" value="create_message" />';
				self::$data['action_create_message_set'] = true;
			}
			$defaultValue = self::GetQuotedSubject($msg, $rev);
		}
		else
		{
			if (!($rev instanceof UnbMessageRevision))
				throw new InvalidArgumentException('Invalid argument type: $rev');

			if (!self::$data['action_alter_message_set'])
			{
				echo '<input type="hidden" name="action[]" value="alter_message" />';
				self::$data['action_alter_message_set'] = true;
			}
			$defaultValue = $rev->GetSubject();
			echo '<input type="hidden" name="message_subject_hash" value="' . md5($defaultValue) . '" />';
		}
		echo '<input type="text" id="' . $id . '" name="message_subject" size="60" ' . self::GetStyleClassAttr($style) . ' value="' . self::GetValue('message_subject', $defaultValue) . '" />';
	}

	// Prints the input element for the message content.
	//
	// newMessage = (bool) {{true}} to create a new message (replying to another one) or {{false}} to alter a message
	// msg = (UnbMessage) Message instance to be replied to resp. to be altered, or {{null}} for an empty new message
	// rev = (UnbMessageRevision) Message revision instance to be replied to resp. to be altered, or {{null}} for an empty new message
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function MsgContent($newMessage = true, $msg = null, $rev = null, $id = null, $style = '')
	{
		$id = self::GetId($id, 'msgContentTb');

		if ($newMessage)
		{
			if (!self::$data['action_create_message_set'])
			{
				echo '<input type="hidden" name="action[]" value="create_message" />';
				self::$data['action_create_message_set'] = true;
			}
			$defaultValue = self::GetQuotedMsg($msg, $rev);
		}
		else
		{
			if (!($rev instanceof UnbMessageRevision))
				throw new InvalidArgumentException('Invalid argument type: $rev');

			if (!self::$data['action_alter_message_set'])
			{
				echo '<input type="hidden" name="action[]" value="alter_message" />';
				self::$data['action_alter_message_set'] = true;
			}
			$defaultValue = $rev->GetContent();
			echo '<input type="hidden" name="message_content_hash" value="' . md5($defaultValue) . '" />';
		}
		echo '<textarea id="' . $id . '" name="message_content" cols="60" rows="15" ' . self::GetStyleClassAttr($style) . '>' . self::GetValue('message_content', $defaultValue) . '</textarea>';
	}

	// Prints the input element for the message change summary.
	//
	// The default value is always empty, because you cannot alter a message revision and a new
	// revision always comes with a new change summary.
	//
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function MsgSummary($id = null, $style = '')
	{
		$id = self::GetId($id, 'msgSummaryTb');

		if (!self::$data['action_alter_message_set'])
		{
			echo '<input type="hidden" name="action[]" value="alter_message" />';
			self::$data['action_alter_message_set'] = true;
		}
		echo '<input type="text" id="' . $id . '" name="message_summary" size="60" ' . self::GetStyleClassAttr($style) . ' value="' . self::GetValue('message_summary') . '" />';
	}

	// Prints the input element for the message file attachment.
	//
	// newMessage = (bool) {{true}} to create a new message (replying to another one) or {{false}} to alter a message
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	//
	public static function MsgAttach($newMessage = true, $id = null)
	{
		$id = self::GetId($id, 'msgFile');

		if ($newMessage)
		{
			if (!self::$data['action_create_message_set'])
			{
				echo '<input type="hidden" name="action[]" value="create_message" />';
				self::$data['action_create_message_set'] = true;
			}
			echo '<input type="file" id="' . $id . '" name="message_filedata" size="50" />
				<input type="hidden" name="set_messagerevision_filedata" value="1" />';
		}
		else
		{
			if (!self::$data['action_alter_message_set'])
			{
				echo '<input type="hidden" name="action[]" value="alter_message" />';
				self::$data['action_alter_message_set'] = true;
			}
			#echo 'Replace/add attached uploaded file: <input type="file" name="message_filedata" size="50" /><br />
			#	<label><input type="checkbox" name="message_cleardata" value="1" ' . ($_POST['message_cleardata'] ? 'checked="checked"' : '') . ' /> Clear attached file</label>
			#	<input type="hidden" name="set_messagerevision_filedata" value="1" />';
			// NOTE: File data cannot be added or removed later, it can only be replaced. A file remains a file, a non-file won't become a file.
			echo '<input type="file" id="' . $id . '" name="message_filedata" size="50" /><br />
				<input type="hidden" name="set_messagerevision_filedata" value="1" />';
		}
	}

	// Prints the checkbox element to remove a file from a message.
	//
	// fileId = (int) Message ID of the file to be removed by ticking the checkbox
	//
	public static function MsgRemoveFile($fileId)
	{
		echo '<label><input type="checkbox" name="message_removefile[]" value="' . intval($fileId) . '" /> ' . UnbLocale::Translate('msg:remove this file') . '</label>';
	}

	// Prints the input element for the message page name.
	//
	// msg = (UnbMessage) Message instance to be altered, or {{null}} for a new message
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function MsgPageName($msg = null, $id = null, $style = '')
	{
		$defaultValue = '';
		if (isset($msg))
		{
			if (!($msg instanceof UnbMessage))
				throw new InvalidArgumentException('Invalid argument type: $msg');
			$defaultValue = $msg->GetPageName();
		}
		$id = self::GetId($id, 'msgPageNameTb');

		echo '<input type="text" id="' . $id . '" name="message_pagename" size="40" ' . self::GetStyleClassAttr($style) . ' value="' . self::GetValue('message_pagename', $defaultValue) . '" />
			<input type="hidden" name="action[]" value="set_message_pagename" />';
	}

	// Prints the input element for the message discussion entry point (DEP) flag.
	//
	// param = (UnbMessage) Message instance to be altered
	//         (bool) Default value for a new message
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	//
	public static function MsgDEP($param, $id = null)
	{
		$defaultValue = false;
		if ($param instanceof UnbMessage)
			$defaultValue = $param->IsEntryPoint();
		else if (is_bool($param))
			$defaultValue = $param;
		else
			throw new InvalidArgumentException('Invalid argument type: $param');
		$id = self::GetId($id, 'msgDEPCb');

		echo '<label><input type="checkbox" id="' . $id . '" name="message_entrypoint" value="1" ' . (($_POST ? $_POST['message_entrypoint'] : $defaultValue) ? 'checked="checked"' : '') . ' /> ' . UnbLocale::Translate('msg:dep option') . '</label>
			<input type="hidden" name="action[]" value="set_message_entrypoint" />';
	}

	// Recursion helper function for MsgTags.
	//
	private static function MsgTags_ListTags($tags, $msgTags, $parent = 0, $level = 0)
	{
		$divOnThisLevel = false;
		foreach ($tags as $tag)
		{
			if ($tag->GetParentTagId() != $parent) continue;

			if ($level > 0 && !$divOnThisLevel)
			{
				echo '<div style="border-left: solid 1px #dddddd; padding-left: 1.5em; margin-top: ' . (0.2 / $level) . 'em; margin-bottom: ' . (0.4 / $level) . 'em;">';
				$divOnThisLevel = true;
			}

			if ($tag->GetFlags() & UnbTagFlag::NotSelectable)
			{
				#echo str_repeat('&nbsp; &nbsp; &nbsp; ', $level);
				#echo '<div style="padding-left: 20px;">' . htmlspecialchars($tag->GetTagName()) . '</div>';
				echo '<span title="' . UnbLocale::Translate('tag:not selectable tip') . '"><input type="checkbox" disabled="disabled" /> ' . htmlspecialchars($tag->GetTagName()) . '</span><br />';
			}
			else if (!$tag->IsUseAllowed())
			{
				echo '<span title="' . UnbLocale::Translate('tag:not allowed tip') . '"><input type="checkbox" disabled="disabled" /> ' . htmlspecialchars($tag->GetTagName()) . '</span><br />';
			}
			else
			{
				$selected = false;
				$userData = '';
				if ($_POST)
				{
					$selected = in_array($tag->GetTagId(), $_POST['message_tagid']);
					$userData = $_POST['message_tagdata_' . $tag->GetTagId()];
				}
				else
				{
					foreach ($msgTags as $msgTag)
					{
						if (is_int($msgTag) && $msgTag == $tag->GetTagId())
						{
							$selected = true;
							break;
						}
						if (is_array($msgTag) && $msgTag['TagId'] == $tag->GetTagId())
						{
							$selected = true;
							$userData = $msgTag['TagData'];
						}
					}
				}

				#echo str_repeat('&nbsp; &nbsp; &nbsp; ', $level);
				echo '<label><input type="checkbox" name="message_tagid[]" value="' . $tag->GetTagId() . '"' . ($selected ? ' checked="checked"' : '') . ' /> ' . htmlspecialchars($tag->GetTagName()) . '</label>';
				if ($tag->GetFlags() & UnbTagFlag::Invisible)
					echo ' [' . UnbLocale::Translate('tag:invisible') . ']';
				if ($tag->GetFlags() & UnbTagFlag::UserData)
				{
					echo ':&nbsp; <input type="text" name="message_tagdata_' . $tag->GetTagId() . '" size="30" value="' . htmlspecialchars($userData) . '" />';
				}
				if (strlen($tag->GetDescription()))
				{
					echo ' &nbsp;–&nbsp; <small>' . htmlspecialchars($tag->GetDescription()) . '</small>';
				}
				echo '<br />';
			}

			self::MsgTags_ListTags($tags, $msgTags, $tag->GetTagId(), $level + 1);
		}

		if ($divOnThisLevel)
		{
			echo '</div>';
		}
	}

	// Prints the list of available tags to be set for the message.
	//
	// rev = (UnbMessageRevision) Message revision instance to be altered or replied to, or {{null}} for an empty selection
	//
	public static function MsgTags($rev = null)
	{
		$msgTags = array();
		if (isset($rev))
		{
			if (!($rev instanceof UnbMessageRevision))
				throw new InvalidArgumentException('Invalid argument type: $rev');
			$msgTags = $rev->GetTagIds();
		}

		$tags = UnbTag::Find();
		self::MsgTags_ListTags($tags, $msgTags);
		echo '<input type="hidden" name="action[]" value="set_messagerevision_tags" />';
	}

	// Prints the input elements for the message moderation flags.
	//
	// msg = (UnbMessage) Message instance to be altered, or {{null}} for a new message
	//
	public static function MsgModerateMsg($msg = null)
	{
		$locked = false;
		$hidden = false;
		$enforceApproval = false;
		if (isset($msg))
		{
			if (!($msg instanceof UnbMessage))
				throw new InvalidArgumentException('Invalid argument type: $msg');
			$locked = $msg->GetModerationFlags() & UnbMessageModerationFlag::Locked;
			$hidden = $msg->GetModerationFlags() & UnbMessageModerationFlag::Hidden;
			$enforceApproval = $msg->GetModerationFlags() & UnbMessageModerationFlag::EnforceApproval;
		}

		echo '<label><input type="checkbox" name="message_locked" value="1" ' . (($_POST ? $_POST['message_locked'] : $locked) ? 'checked="checked"' : '') . ' /> ' . UnbLocale::Translate('msg:lock this message') . '</label><br />
			<label><input type="checkbox" name="message_hidden" value="1" ' . (($_POST ? $_POST['message_hidden'] : $hidden) ? 'checked="checked"' : '') . ' /> ' . UnbLocale::Translate('msg:hide this message') . '</label><br />
			<label><input type="checkbox" name="message_enforce_approval" value="1" ' . (($_POST ? $_POST['message_enforce_approval'] : $enforceApproval) ? 'checked="checked"' : '') . ' /> ' . UnbLocale::Translate('msg:enforce approval') . '</label><br />
			<input type="hidden" name="action[]" value="moderate_message" />';
	}

	// Gets a message's read access keys as multi-line string.
	//
	// msg = (UnbMessage) Message instance to get the key list from
	//
	// Returns (string) All read access keys with their display names, one per line, separated by \n
	//
	public static function GetMsgReadAccessKeys($msg)
	{
		if (!($msg instanceof UnbMessage))
			throw new InvalidArgumentException('Invalid argument type: $msg');

		return self::GetUserList($msg->GetReadAccessKeys());
	}

	// Prints the input element for the message read access list.
	//
	// msg = (UnbMessage) Message instance to be altered, or {{null}} for a new message
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function MsgReadAccessList($msg = null, $id = null, $style = '')
	{
		$readAccessList = '';
		if (isset($msg))
		{
			if (!($msg instanceof UnbMessage))
				throw new InvalidArgumentException('Invalid argument type: $msg');
			$readAccessList = self::GetMsgReadAccessKeys($msg);
		}
		$id = self::GetId($id, 'msgReadAccessListTb');

		echo '<textarea id="' . $id . '" name="message_readaccesslist" cols="30" rows="3" ' . self::GetStyleClassAttr($style) . '>' . self::GetValue('message_readaccesslist', $readAccessList, 'message_readaccess_empty') . '</textarea>
			<input type="hidden" name="message_readaccess_empty" value="0" />
			<input type="hidden" name="action[]" value="set_message_readaccess" />';

		$defaultMessage = UnbLocale::Translate('access:default everybody click to add');
		if (strlen($defaultMessage))
			self::InsertDefaultMessage($id, $defaultMessage, 'message_readaccess_empty');
	}

	// Gets a message's alter access keys as multi-line string.
	//
	// msg = (UnbMessage) Message instance to get the key list from
	//
	// Returns (string) All alter access keys with their display names, one per line, separated by \n
	//
	public static function GetMsgAlterAccessKeys($msg)
	{
		if (!($msg instanceof UnbMessage))
			throw new InvalidArgumentException('Invalid argument type: $msg');

		return self::GetUserList($msg->GetAlterAccessKeys());
	}

	// Prints the input element for the message alter access list.
	//
	// msg = (UnbMessage) Message instance to be altered, or {{null}} for a new message with an empty list
	//       (array(int)) Default list of keys
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function MsgAlterAccessList($msg = null, $id = null, $style = '')
	{
		$alterAccessList = '';
		if (isset($msg))
		{
			if ($msg instanceof UnbMessage)
				$alterAccessList = self::GetMsgAlterAccessKeys($msg);
			else if (is_array($msg))
				$alterAccessList = self::GetUserList($msg);
			else
				throw new InvalidArgumentException('Invalid argument type: $msg');
		}
		$id = self::GetId($id, 'msgAlterAccessListTb');

		echo '<textarea id="' . $id . '" name="message_alteraccesslist" cols="30" rows="3" ' . self::GetStyleClassAttr($style) . '>' . self::GetValue('message_alteraccesslist', $alterAccessList, 'message_alteraccess_empty') . '</textarea>
			<input type="hidden" name="message_alteraccess_empty" value="0" />
			<input type="hidden" name="action[]" value="set_message_alteraccess" />';

		$defaultMessage = UnbLocale::Translate('access:default everybody click to add');
		if (strlen($defaultMessage))
			self::InsertDefaultMessage($id, $defaultMessage, 'message_alteraccess_empty');
	}

	// Gets a message's reply access keys as multi-line string.
	//
	// msg = (UnbMessage) Message instance to get the key list from
	//
	// Returns (string) All reply access keys with their display names, one per line, separated by \n
	//
	public static function GetMsgReplyAccessKeys($msg)
	{
		if (!($msg instanceof UnbMessage))
			throw new InvalidArgumentException('Invalid argument type: $msg');

		return self::GetUserList($msg->GetReplyAccessKeys());
	}

	// Prints the input element for the message reply access list.
	//
	// msg = (UnbMessage) Message instance to be altered, or {{null}} for a new message
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function MsgReplyAccessList($msg = null, $id = null, $style = '')
	{
		$replyAccessList = '';
		if (isset($msg))
		{
			if (!($msg instanceof UnbMessage))
				throw new InvalidArgumentException('Invalid argument type: $msg');
			$replyAccessList = self::GetMsgReplyAccessKeys($msg);
		}
		$id = self::GetId($id, 'msgReplyAccessListTb');

		echo '<textarea id="' . $id . '" name="message_replyaccesslist" cols="30" rows="3" ' . self::GetStyleClassAttr($style) . '>' . self::GetValue('message_replyaccesslist', $replyAccessList, 'message_replyaccess_empty') . '</textarea>
			<input type="hidden" name="message_replyaccess_empty" value="0" />
			<input type="hidden" name="action[]" value="set_message_replyaccess" />';

		$defaultMessage = UnbLocale::Translate('access:default everybody click to add');
		if (strlen($defaultMessage))
			self::InsertDefaultMessage($id, $defaultMessage, 'message_replyaccess_empty');
	}

	// ---------- Message editor style button display functions ----------

	// Prints the input button to emphasise text in the message content text field.
	//
	// id = (string) HTML element ID of the content input element, or {{null}} for the default ID
	// text = (string) Button text (will be HTML-encoded), or {{null}} for built-in default text
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ButtonEm($textboxId = null, $text = null, $style = '')
	{
		$textboxId = self::GetId($textboxId, 'msgContentTb');
		if (!isset($text))
			$text = UnbLocale::Translate('format:emphasis');

		echo '<input type="button" value="' . htmlspecialchars($text) . '" ' . self::GetStyleClassAttr($style, 'font-style: italic;|icon13 em') . ' onclick="PlainTypeEm(\'' . $textboxId . '\')" />';
	}

	// Prints the input button to strongly emphasise text in the message content text field.
	//
	// id = (string) HTML element ID of the content input element, or {{null}} for the default ID
	// text = (string) Button text (will be HTML-encoded), or {{null}} for built-in default text
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ButtonStrong($textboxId = null, $text = null, $style = '')
	{
		$textboxId = self::GetId($textboxId, 'msgContentTb');
		if (!isset($text))
			$text = UnbLocale::Translate('format:strong emphasis');

		echo '<input type="button" value="' . htmlspecialchars($text) . '" ' . self::GetStyleClassAttr($style, 'font-weight: bold;|icon13 strong') . ' onclick="PlainTypeStrong(\'' . $textboxId . '\')" />';
	}

	// Prints the input button to quote text in the message content text field.
	//
	// id = (string) HTML element ID of the content input element, or {{null}} for the default ID
	// text = (string) Button text (will be HTML-encoded), or {{null}} for built-in default text
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ButtonQuote($textboxId = null, $text = null, $style = '')
	{
		$textboxId = self::GetId($textboxId, 'msgContentTb');
		if (!isset($text))
			$text = UnbLocale::Translate('format:quotation');

		echo '<input type="button" value="' . htmlspecialchars($text) . '" ' . self::GetStyleClassAttr($style, '|icon13 quote') . ' onclick="PlainTypeQuote(\'' . $textboxId . '\')" />';
	}

	// Prints the input button to indent text in the message content text field.
	//
	// id = (string) HTML element ID of the content input element, or {{null}} for the default ID
	// text = (string) Button text (will be HTML-encoded), or {{null}} for built-in default text
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ButtonIndent($textboxId = null, $text = null, $style = '')
	{
		$textboxId = self::GetId($textboxId, 'msgContentTb');
		if (!isset($text))
			$text = UnbLocale::Translate('format:indentation');

		echo '<input type="button" value="' . htmlspecialchars($text) . '" ' . self::GetStyleClassAttr($style, '|icon13 indent') . ' onclick="PlainTypeIndent(\'' . $textboxId . '\')" />';
	}

	// Prints the input button to make an unordered list in the message content text field.
	//
	// id = (string) HTML element ID of the content input element, or {{null}} for the default ID
	// text = (string) Button text (will be HTML-encoded), or {{null}} for built-in default text
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ButtonUl($textboxId = null, $text = null, $style = '')
	{
		$textboxId = self::GetId($textboxId, 'msgContentTb');
		if (!isset($text))
			$text = UnbLocale::Translate('format:list');

		echo '<input type="button" value="' . htmlspecialchars($text) . '" ' . self::GetStyleClassAttr($style, '|icon13 ul') . ' onclick="PlainTypeUList(\'' . $textboxId . '\')" />';
	}

	// Prints the input button to make an ordered list in the message content text field.
	//
	// id = (string) HTML element ID of the content input element, or {{null}} for the default ID
	// text = (string) Button text (will be HTML-encoded), or {{null}} for built-in default text
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ButtonOl($textboxId = null, $text = null, $style = '')
	{
		$textboxId = self::GetId($textboxId, 'msgContentTb');
		if (!isset($text))
			$text = UnbLocale::Translate('format:ordered list');

		echo '<input type="button" value="' . htmlspecialchars($text) . '" ' . self::GetStyleClassAttr($style, '|icon13 ol') . ' onclick="PlainTypeOList(\'' . $textboxId . '\')" />';
	}

	// Prints the input button to mark text as code in the message content text field.
	//
	// id = (string) HTML element ID of the content input element, or {{null}} for the default ID
	// text = (string) Button text (will be HTML-encoded), or {{null}} for built-in default text
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ButtonCode($textboxId = null, $text = null, $style = '')
	{
		$textboxId = self::GetId($textboxId, 'msgContentTb');
		if (!isset($text))
			$text = UnbLocale::Translate('format:code');

		echo '<input type="button" value="' . htmlspecialchars($text) . '" ' . self::GetStyleClassAttr($style, '|icon13 code') . ' onclick="PlainTypeCode(\'' . $textboxId . '\')" />';
	}

	// Prints the input button to strike out text in the message content text field.
	//
	// id = (string) HTML element ID of the content input element, or {{null}} for the default ID
	// text = (string) Button text (will be HTML-encoded), or {{null}} for built-in default text
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ButtonStrike($textboxId = null, $text = null, $style = '')
	{
		$textboxId = self::GetId($textboxId, 'msgContentTb');
		if (!isset($text))
			$text = UnbLocale::Translate('format:strike');

		echo '<input type="button" value="' . htmlspecialchars($text) . '" ' . self::GetStyleClassAttr($style, 'font-decoration: line-through;|icon13 strike') . ' onclick="PlainTypeStrike(\'' . $textboxId . '\')" />';
	}

	// Prints the input button to add a link in the message content text field.
	//
	// id = (string) HTML element ID of the content input element, or {{null}} for the default ID
	// text = (string) Button text (will be HTML-encoded), or {{null}} for built-in default text
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ButtonLink($textboxId = null, $text = null, $style = '')
	{
		$textboxId = self::GetId($textboxId, 'msgContentTb');
		if (!isset($text))
			$text = UnbLocale::Translate('format:web link');

		echo '<input type="button" value="' . htmlspecialchars($text) . '" ' . self::GetStyleClassAttr($style, '|icon13 link') . ' onclick="PlainTypeLink(\'' . $textboxId . '\')" />';
	}

	// Prints all input buttons to format text in the message content text field.
	//
	// id = (string) HTML element ID of the content input element, or {{null}} for the default ID
	//
	public static function Buttons($textboxId = null)
	{
		echo '<em>' . UnbLocale::Translate('msg:text formatting header') . UnbLocale::Colon() . '</em>';
		echo '<div style="line-height: 200%; margin: 0.5em 0 1em;">';
		self::ButtonEm($textboxId);
		echo ' ';
		self::ButtonStrong($textboxId);
		echo '<br />';
		self::ButtonQuote($textboxId);
		echo ' ';
		self::ButtonIndent($textboxId);
		echo '<br />';
		self::ButtonUl($textboxId);
		echo ' ';
		self::ButtonOl($textboxId);
		echo '<br />';
		self::ButtonCode($textboxId);
		echo ' ';
		self::ButtonStrike($textboxId);
		echo '<br />';
		self::ButtonLink($textboxId);
		echo '</div>';
		echo '<a href="' . self::$data['plaintype_help_url'] . '" onclick="return openPopupLink(this, \'plaintypehelp\', ' . self::$data['plaintype_help_width'] . ', ' . self::$data['plaintype_help_height'] . ')">' . UnbLocale::Translate('msg:formatting learn mode') . '</a>';
	}

	// ---------- Tag management display functions ----------

	// Prints the input element for the tag name.
	//
	// tag = (UnbTag) Tag instance to be altered, or {{null}} for a new tag
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function TagName($tag = null, $id = null, $style = '')
	{
		$defaultValue = '';
		if (isset($tag))
		{
			if (!($tag instanceof UnbTag))
				throw new InvalidArgumentException('Invalid argument type: $tag');
			$defaultValue = $tag->GetTagName();
		}
		$id = self::GetId($id, 'tagNameTb');

		echo '<input type="text" id="' . $id . '" name="tag_name" size="30" ' . self::GetStyleClassAttr($style) . ' value="' . self::GetValue('tag_name', $defaultValue) . '" />';

		if (!self::IsStyleDisabled($style))
			echo '<input type="hidden" name="action[]" value="set_tag_name" />';
	}

	// Prints the input element for the tag description.
	//
	// tag = (UnbTag) Tag instance to be altered, or {{null}} for a new tag
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function TagDescription($tag = null, $id = null, $style = '')
	{
		$defaultValue = '';
		if (isset($tag))
		{
			if (!($tag instanceof UnbTag))
				throw new InvalidArgumentException('Invalid argument type: $tag');
			$defaultValue = $tag->GetDescription();
		}
		$id = self::GetId($id, 'tagDescriptionTb');

		echo '<input type="text" id="' . $id . '" name="tag_description" size="50" ' . self::GetStyleClassAttr($style) . ' value="' . self::GetValue('tag_description', $defaultValue) . '" />';

		if (!self::IsStyleDisabled($style))
			echo '<input type="hidden" name="action[]" value="set_tag_description" />';
	}

	// Prints the input elements for the tag flags.
	//
	// tag = (UnbTag) Tag instance to be altered, or {{null}} for a new tag
	// disabled = (bool) Keep the user from changing the values
	//
	public static function TagFlags($tag = null, $disabled = false)
	{
		$notSelectable = false;
		$invisible = false;
		$enforceApproval = false;
		$userData = false;
		if (isset($tag))
		{
			if (!($tag instanceof UnbTag))
				throw new InvalidArgumentException('Invalid argument type: $tag');
			$notSelectable = $tag->GetFlags() & UnbTagFlag::NotSelectable;
			$invisible = $tag->GetFlags() & UnbTagFlag::Invisible;
			$enforceApproval = $tag->GetFlags() & UnbTagFlag::EnforceApproval;
			$userData = $tag->GetFlags() & UnbTagFlag::UserData;
		}

		$dis = $disabled ? ' disabled="disabled"' : '';

		echo '<label><input type="checkbox" name="tag_flag_notselectable" value="1" ' . (($_POST ? $_POST['tag_flag_notselectable'] : $notSelectable) ? 'checked="checked"' : '') . $dis . ' /> ' . UnbLocale::Capital(UnbLocale::Translate('tag:not selectable')) . '</label><br />
			<label><input type="checkbox" name="tag_flag_invisible" value="1" ' . (($_POST ? $_POST['tag_flag_invisible'] : $invisible) ? 'checked="checked"' : '') . $dis . ' /> ' . UnbLocale::Capital(UnbLocale::Translate('tag:invisible')) . '</label><br />
			<label><input type="checkbox" name="tag_flag_enforceapproval" value="1" ' . (($_POST ? $_POST['tag_flag_enforceapproval'] : $enforceApproval) ? 'checked="checked"' : '') . $dis . ' /> ' . UnbLocale::Capital(UnbLocale::Translate('tag:enforce approval')) . '</label><br />
			<label><input type="checkbox" name="tag_flag_userdata" value="1" ' . (($_POST ? $_POST['tag_flag_userdata'] : $userData) ? 'checked="checked"' : '') . $dis . ' /> ' . UnbLocale::Capital(UnbLocale::Translate('tag:user data')) . '</label><br />';

		if (!$disabled)
			echo '<input type="hidden" name="action[]" value="set_tag_flags" />';
	}

	// Prints the input element for the parent tag selection.
	//
	// tag = (UnbTag) Tag instance to be altered, or {{null}} for a new tag
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function ParentTag($tag = null, $id = null, $style = '')
	{
		$defaultValue = 0;
		if (isset($tag))
		{
			if (!($tag instanceof UnbTag))
				throw new InvalidArgumentException('Invalid argument type: $tag');
			$defaultValue = $tag->GetParentTagId();
		}
		$id = self::GetId($id, 'parentTagLb');

		echo '<select id="' . $id . '" name="tag_parentid" ' . self::GetStyleClassAttr($style) . '>
			<option value="0">' . UnbLocale::Parentheses(UnbLocale::Capital(UnbLocale::Translate('tag:none'))) . '</option>';
		$tags = UnbTag::Find();
		$levels = array();
		foreach ($tags as $tag2)
		{
			$level = $levels[$tag2->GetParentTagId()];
			if (!isset($level))
				$level = 0;   // No parent level found, start at 0
			else
				$level++;   // One level deeper than parent
			$levels[$tag2->GetTagId()] = $level;

			if ($tag2->GetFlags() & UnbTagFlag::Invisible) continue;
			#if ($tag2->GetFlags() & UnbTagFlag::UserData) continue;
			echo '<option value="' . $tag2->GetTagId() . '"';
			if ($_POST && $tag2->GetTagId() == $_POST['tag_parentid'] ||
				!$_POST && $tag2->GetTagId() == $defaultValue)
				echo ' selected="selected"';
			echo '>' .
				str_repeat('&nbsp; &nbsp; ', $level) . htmlspecialchars($tag2->GetTagName()) . /*' (' . $tag2->GetTagId() . ')' .*/ '</option>';
		}
		echo '</select>';

		if (!self::IsStyleDisabled($style))
			echo '<input type="hidden" name="action[]" value="set_tag_parent" />';
	}

	// Gets a tag's read access keys as multi-line string.
	//
	// tag = (UnbTag) Tag instance to get the key list from
	//
	// Returns (string) All read access keys with their display names, one per line, separated by \n
	//
	public static function GetTagReadAccessKeys($tag)
	{
		if (!($tag instanceof UnbTag))
			throw new InvalidArgumentException('Invalid argument type: $tag');

		return self::GetUserList($tag->GetReadAccessKeys());
	}

	// Prints the input element for the tag read access list.
	//
	// tag = (UnbTag) Tag instance to be altered, or {{null}} for a new tag
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function TagReadAccessList($tag = null, $id = null, $style = '')
	{
		$readAccessList = '';
		if (isset($tag))
		{
			if (!($tag instanceof UnbTag))
				throw new InvalidArgumentException('Invalid argument type: $tag');
			$readAccessList = self::GetTagReadAccessKeys($tag);
		}
		$id = self::GetId($id, 'tagReadAccessListTb');

		echo '<textarea id="' . $id . '" name="tag_readaccesslist" cols="30" rows="3" ' . self::GetStyleClassAttr($style) . '>' . self::GetValue('tag_readaccesslist', $readAccessList, 'tag_readaccess_empty') . '</textarea>
			<input type="hidden" name="tag_readaccess_empty" value="0" />';

		if (!self::IsStyleDisabled($style))
		{
			echo '<input type="hidden" name="action[]" value="set_tag_readaccess" />';
			$defaultMessage = UnbLocale::Translate('access:default everybody click to add');
		}
		else
		{
			$defaultMessage = UnbLocale::Translate('access:default everybody');
		}
		if (strlen($defaultMessage))
			self::InsertDefaultMessage($id, $defaultMessage, 'tag_readaccess_empty');
	}

	// Gets a tag's use access keys as multi-line string.
	//
	// tag = (UnbTag) Tag instance to get the key list from
	//
	// Returns (string) All use access keys with their display names, one per line, separated by \n
	//
	public static function GetTagUseAccessKeys($tag)
	{
		if (!($tag instanceof UnbTag))
			throw new InvalidArgumentException('Invalid argument type: $tag');

		return self::GetUserList($tag->GetUseAccessKeys());
	}

	// Prints the input element for the tag use access list.
	//
	// tag = (UnbTag) Tag instance to be altered, or {{null}} for a new tag
	// id = (string) HTML element ID for the input element, or {{null}} for a default ID
	// style = (string) HTML element style attribute value, see GetStyleClassAttr
	//
	public static function TagUseAccessList($tag = null, $id = null, $style = '')
	{
		$useAccessList = '';
		if (isset($tag))
		{
			if (!($tag instanceof UnbTag))
				throw new InvalidArgumentException('Invalid argument type: $tag');
			$useAccessList = self::GetTagUseAccessKeys($tag);
		}
		$id = self::GetId($id, 'tagUseAccessListTb');

		echo '<textarea id="' . $id . '" name="tag_useaccesslist" cols="30" rows="3" ' . self::GetStyleClassAttr($style) . '>' . self::GetValue('tag_useaccesslist', $useAccessList, 'tag_useaccess_empty') . '</textarea>
			<input type="hidden" name="tag_useaccess_empty" value="0" />';

		if (!self::IsStyleDisabled($style))
		{
			echo '<input type="hidden" name="action[]" value="set_tag_useaccess" />';
			$defaultMessage = UnbLocale::Translate('access:default everybody click to add');
		}
		else
		{
			$defaultMessage = UnbLocale::Translate('access:default everybody');
		}
		if (strlen($defaultMessage))
			self::InsertDefaultMessage($id, $defaultMessage, 'tag_useaccess_empty');
	}

	// Prints an image tag for a file content type icon.
	//
	// contentType = (string) File content type
	// ret = (bool) Specifies whether the HTML code shall be returned instead of printed
	//
	public static function ContentTypeIcon($contentType, $ret = false)
	{
		$path = self::$data['contenttype_img_url'];
		$fileName = str_replace('/', '_', $contentType);
		if (!file_exists($path . $fileName . '.png'))
			$fileName = preg_replace('~/.*$~', '', $contentType);
		if (!file_exists($path . $fileName . '.png'))
			$fileName = 'unknown';

		$html = '<img src="' . $path . $fileName . '.png" alt="' . htmlspecialchars($contentType) . '" style="vertical-align: middle; margin: -2px 5px 2px 0;" />';
		if ($ret)
			return $html;
		else
			echo $html;
	}

} // class UnbUi

?>
